# coding: utf-8

from time import time, sleep
from numpy import uint8, ndarray, uint16, copy, squeeze
from typing import Tuple, Optional, Union, List
from re import findall, search
from subprocess import Popen, PIPE, run
from fractions import Fraction
from platform import system
import logging

from .meta_camera import Camera
from .._global import OptionalModule

try:
  import cv2
except (ImportError, ModuleNotFoundError):
  pass
  cv2 = OptionalModule('opencv-python')

try:
  from gi import require_version
  require_version('Gst', '1.0')
  require_version('GstApp', '1.0')
  from gi.repository import Gst, GstApp
except (ImportError, ModuleNotFoundError, ValueError):
  Gst = GstApp = OptionalModule('PyGObject')


class CameraGstreamer(Camera):
  """A class for reading images from a video device using Gstreamer.

  It can read images from the default video source, or a video device can be
  specified. The user has access to a range of parameters for tuning the image,
  depending on the capability of the hardware. Alternatively, it is possible to
  give a custom GStreamer pipeline as an argument. In this case no settings are
  available, and it is up to the user to ensure the validity of the pipeline.

  This class uses less resources and is compatible with more cameras than the
  :class:`~crappy.camera.CameraOpencv` camera, that relies on OpenCV. The
  installation of GStreamer is however less straightforward than the one of
  OpenCV, especially on Windows !

  Warning:
    There are two classes for CameraGstreamer, one for Linux based on
    `v4l-utils`, and another one for Linux (without `v4l-utils`) and other OS.
    Depending on the installation of `v4l-utils` and the OS, the correct class
    will be automatically imported. The version using `v4l-utils` allows tuning
    more parameters than the basic version.

  Note:
    This Camera requires the module :mod:`PyGObject` to be installed, as well
    as GStreamer.
  
  .. versionadded:: 1.5.9
  .. versionchanged:: 2.0.0 renamed from Camera_gstreamer to CameraGstreamer
  """

  def __init__(self) -> None:
    """Simply initializes the instance attributes."""

    super().__init__()

    Gst.init(None)
    self._last_frame_nr = 0
    self._frame_nr = 0

    # These attributes will be set later
    self._pipeline = None
    self._process: Optional[Popen] = None
    self._img: Optional[ndarray] = None
    self._device: Optional[Union[str, int]] = None
    self._user_pipeline: Optional[str] = None
    self._nb_channels: int = 3
    self._img_depth: int = 8
    self._formats: List[str] = list()
    self._app_sink = None

  def open(self,
           device: Optional[Union[int, str]] = None,
           user_pipeline: Optional[str] = None,
           nb_channels: Optional[int] = None,
           img_depth: Optional[int] = None,
           **kwargs) -> None:
    """Opens the pipeline, sets the settings and starts the acquisition of
    images.

    A custom pipeline can be specified using the ``user_pipeline`` argument.
    Simply copy-paste the pipeline as it is in the terminal, and the class will
    take care of adjusting it to make it compatible with the Python binding
    of GStreamer.

    Note:
      For custom pipelines, it may be necessary to set a format directly in the
      pipeline (which isn't the case in the terminal). For instance, this line

      .. code-block:: shell-session

        gst-launch-1.0 autovideosrc ! videoconvert ! autovideosink

      May need to become

      .. code-block:: shell-session

        gst-launch-1.0 autovideosrc ! videoconvert ! video/x-raw,format=BGR ! \
autovideosink

    Note:
      Pipelines built using a pipe for redirecting the stream from another
      application are also accepted. For example, the following user pipeline
      is valid :

      .. code-block:: shell-session

        gphoto2 --stdout --capture-movie | gst-launch-1.0 fdsrc fd=0 ! \
videoconvert ! autovideosink

    Args:
      device: The video device to open, if the device opened by default isn't
        the right one. In Linux, should be a path like `/dev/video0`. In
        Windows and Mac, should be the index of the video device. This argument
        is ignored if a ``user_pipeline`` is given.
      user_pipeline: A custom pipeline that can optionally be given as a
        :obj:`str`. If given, the ``device`` argument is ignored. The pipeline
        should be given as it would be in a terminal.
      nb_channels: The number of channels expected in the acquired images, in
        case a custom pipeline is given. Otherwise, this argument is ignored.
        For now, Crappy only manages 1- and 3-channel images.
      img_depth: The bit depth of each channel of the acquired images, in case
        a custom pipeline is given. Otherwise, this argument is ignored. For
        now, Crappy only manages 8- and 16-bits deep images.
      **kwargs: Allows specifying values for the settings even before
        displaying the configuration window.
    """

    # Checking the validity of the arguments
    if device is not None and system() == 'Linux' and not isinstance(device,
                                                                     str):
      raise ValueError("In Linux, device should be a string !")
    elif device is not None and system() in ['Darwin', 'Windows'] and not \
            isinstance(device, int):
      raise ValueError("In Windows and Mac, device should be an integer !")

    if img_depth is not None and img_depth not in [8, 16]:
      raise ValueError('The img_depth must be either 8 or 16 (bits)')

    if user_pipeline is not None and nb_channels is None:
      raise ValueError('nb_channels must be given if user_pipeline is !')
    if user_pipeline is not None and img_depth is None:
      raise ValueError('img_depth must be given if user_pipeline is !')

    # Parsing the user pipeline if given
    if user_pipeline is not None:

      # Removing the call to gst-launch-1.0
      user_pipeline = user_pipeline.replace('gst-launch-1.0 ', '')
      # Splitting in case there's a pipe in the command
      user_pipeline = user_pipeline.split(' | ', 1)

      # There's no pipe in the pipeline
      if len(user_pipeline) == 1:
        user_pipeline = user_pipeline[0]

        # Setting the sink
        user_pipeline = user_pipeline.split(' ! ')
        user_pipeline[-1] = 'appsink name=sink'
        user_pipeline = ' ! '.join(user_pipeline)

      # There's a pipe in the pipeline
      else:
        user_application = user_pipeline[0]
        user_pipeline = user_pipeline[1]

        # Opening a subprocess handling the first half of the pipeline
        self.log(logging.INFO, f"Running command "
                               f"{user_application.split(' ')}")
        self._process = Popen(user_application.split(' '), stdout=PIPE)
        file_descriptor = self._process.stdout.fileno()

        # Setting the source and the sink, according to the file descriptor of
        # the subprocess
        user_pipeline = user_pipeline.split(' ! ')
        user_pipeline[0] = f'fdsrc fd={file_descriptor}'
        user_pipeline[-1] = 'appsink name=sink'
        user_pipeline = ' ! '.join(user_pipeline)

    # Setting the expected properties of the image
    self._device = device
    self._user_pipeline = user_pipeline
    self._nb_channels = nb_channels if nb_channels is not None else 3
    self._img_depth = img_depth if img_depth is not None else 8

    # Defining the settings in case no custom pipeline was given
    if user_pipeline is None:

      # Checking the available formats with the Gst device monitor
      cam = None
      device_monitor = Gst.DeviceMonitor.new()
      device_monitor.start()
      devices = [device for device in device_monitor.get_devices()
                 if device.get_device_class() == "Video/Source"]

      # Stop right here if there is no connected camera
      if not devices:
        self.log(logging.ERROR, "No camera devices available for reading !")
        raise IOError

      # Finding the specified device in the available ones
      if self._device is not None:
        for device in devices:
          properties = device.get_properties()
          api = properties.get_value('device.api')
          dev = properties.get_value('object.path')
          if dev == f'{api}:{self._device}':
            cam = device
            break

      # Raising an exception if the specified device cannot be found
      if self._device is not None and cam is None:
        self.log(logging.ERROR, f"Could not open the specified camera "
                                f"at: {self._device}")
        raise IOError

      # Defining a default camera if the device is not specified
      if self._device is None:
        cam = devices[0]
        path = cam.get_properties().get_value('object.path')
        self.log(logging.WARNING, f"No specific camera to open specified, "
                                  f"opening {path.split(':')[1]} by default")

      # Getting the available image formats
      caps = cam.get_caps()
      nb_formats = caps.get_size()
      for i in range(nb_formats):
        form = None
        struct = caps.get_structure(i)
        name = struct.get_name()
        if name == 'image/jpeg':
          form = 'MJPG'
        elif name == 'video/x-h264':
          form = 'H264'
        elif name == 'video/x-hevc':
          form = 'HEVC'
        elif name == 'video/x-raw':
          form = struct.get_value('format')
        else:
          self.log(logging.WARNING, f"Found unsupported cap format: {name}, "
                                    f"ignoring it")

        # Formatting the format into the v4l2 syntax
        if form is not None:
          width = struct.get_value('width')
          height = struct.get_value('height')
          framerates = findall(r'(\d+/\d+)', struct.to_string())
          for fps_str in framerates:
            fps = f"{float(Fraction(fps_str)):.3f}"
            self._formats.append(f"{form} {width}x{height} ({fps} fps)")

      device_monitor.stop()

      # Checking if the formats are supported with the installed libraries
      if self._formats:
        h264_available = bool(run(['gst-inspect-1.0', 'avdec_h264'],
                                  capture_output=True, text=True).stdout)
        h265_available = bool(run(['gst-inspect-1.0', 'avdec_h265'],
                                  capture_output=True, text=True).stdout)
        if not h264_available and any(form.split()[0] == 'H264'
                                      for form in self._formats):
          self._formats = [form for form in self._formats
                           if form.split()[0] != 'H264']
          self.log(logging.WARNING, "The video format H264 could be available "
                                    "for the selected camera if "
                                    "gstreamer1.0-libav was installed !")
        if not h265_available and any(form.split()[0] == 'HEVC'
                                      for form in self._formats):
          self._formats = [form for form in self._formats
                           if form.split()[0] != 'HEVC']
          self.log(logging.WARNING, "The video format H265 could be available "
                                    "for the selected camera if "
                                    "gstreamer1.0-libav was installed !")

      if self._formats:
        # Creating the format parameter
        self.add_choice_setting(name='format',
                                choices=tuple(self._formats),
                                setter=self._set_format)

      # These settings are always available no matter the platform
      if self._nb_channels > 1:
        self.add_choice_setting(name="channels", choices=('1', '3'),
                                default='1')
      self.add_scale_setting(name='brightness', lowest=-1., highest=1.,
                             setter=self._set_brightness, default=0.)
      self.add_scale_setting(name='contrast', lowest=0., highest=2.,
                             setter=self._set_contrast, default=1.)
      self.add_scale_setting(name='hue', lowest=-1., highest=1.,
                             setter=self._set_hue, default=0.)
      self.add_scale_setting(name='saturation', lowest=0., highest=2.,
                             setter=self._set_saturation, default=1.)

    # Setting up GStreamer and the callback
    self.log(logging.INFO, "Initializing the GST pipeline")
    self.log(logging.DEBUG, f"The pipeline is {self._get_pipeline()}")
    self._pipeline = Gst.parse_launch(self._get_pipeline())
    self._app_sink = self._pipeline.get_by_name('sink')
    self._app_sink.set_property("emit-signals", True)
    self._app_sink.connect("new-sample", self._on_new_sample)

    # Starting image acquisition
    self.log(logging.INFO, "Starting the GST pipeline")
    self._pipeline.set_state(Gst.State.PLAYING)

    # Checking that images are read as expected
    t0 = time()
    while not self._frame_nr:
      if time() - t0 > 2:
        raise TimeoutError(
          "Waited too long for the first image ! There is probably an error "
          "in the pipeline. The format of the received images may be "
          "unexpected, in which case you can try specifying it.")
      sleep(0.01)

    # Setting the kwargs if any
    self.set_all(**kwargs)

  def get_image(self) -> Tuple[float, ndarray]:
    """Reads the last image acquired from the camera.

    Returns:
      The acquired image, along with a timestamp.
    """

    # Assuming an image rate greater than 0.5 FPS
    # Checking that we don't return the same image twice
    t0 = time()
    while self._last_frame_nr == self._frame_nr:
      if time() - t0 > 2:
        raise TimeoutError("Waited too long for the next image !")
      sleep(0.01)

    self._last_frame_nr = self._frame_nr

    return time(), self.apply_soft_roi(copy(self._img))

  def close(self) -> None:
    """Simply stops the image acquisition."""

    if self._pipeline is not None:
      self.log(logging.INFO, "Stopping the GST pipeline")
      self._pipeline.set_state(Gst.State.NULL)

    # Closes the subprocess started in case a user pipeline containing a pipe
    # was given
    if self._process is not None:
      self.log(logging.INFO, "Stopping the image generating process")
      self._process.terminate()

  def _restart_pipeline(self, pipeline: str) -> None:
    """Stops the current pipeline, redefines it, and restarts it.

    Args:
      pipeline: The new pipeline to use, as a :obj:`str`.
    """

    # Stops the previous pipeline
    self.log(logging.INFO, "Stopping the GST pipeline")
    self._pipeline.set_state(Gst.State.NULL)

    # Redefines the pipeline and the callbacks
    self.log(logging.INFO, "Initializing the GST pipeline")
    self.log(logging.DEBUG, f"The new pipeline is {pipeline}")
    self._pipeline = Gst.parse_launch(pipeline)
    self._app_sink = self._pipeline.get_by_name('sink')
    self._app_sink.set_property("emit-signals", True)
    self._app_sink.connect("new-sample", self._on_new_sample)

    # Restarts the pipeline
    self.log(logging.INFO, "Starting the GST pipeline")
    self._pipeline.set_state(Gst.State.PLAYING)

  def _get_pipeline(self,
                    brightness: Optional[float] = None,
                    contrast: Optional[float] = None,
                    hue: Optional[float] = None,
                    saturation: Optional[float] = None,
                    img_format: Optional[str] = None) -> str:
    """Method that generates a pipeline, according to the given settings.

    If a user-defined pipeline was given, it will always be returned.

    Args:
      brightness: The brightness value to set, as a :obj:`float` between -1 and
        1.
      contrast: The contrast value to set, as a :obj:`float` between 0 and 2.
      hue: The hue value to set, as a :obj:`float` between -1 and 1.
      saturation: The saturation value to set, as a :obj:`float` between 0 and
        2.
      img_format: The image format to set, as a :obj:`str` containing the name
        of the encoding and optionally both the width and height in pixels.

    Returns:
      A pipeline matching the current settings values.
    """

    # Returns the custom pipeline if any
    if self._user_pipeline is not None:
      return self._user_pipeline

    # Choosing the source according to the platform
    if system() == "Linux":
      source = 'v4l2src'
    elif system() == 'Windows':
      source = 'ksvideosrc'
    elif system() == 'Darwin':
      source = 'vfsvideosrc'
    else:
      source = 'autovideosrc'

    # The source argument is handled differently according to the platform
    if system() == 'Linux' and self._device is not None:
      device = f'device={self._device}'
    elif system() in ['Darwin', 'Windows'] and self._device is not None:
      device = f'device-index={self._device}'
    else:
      device = ''

    # Getting the format string if not provided as an argument
    if img_format is None and hasattr(self, 'format'):
      img_format = self.format
    # In case it goes wrong, just try without specifying the format
    if img_format is None:
      img_format = ''

    try:
      format_name, img_size, fps = findall(r"(\w+)\s(\w+)\s\((\d+.\d+) fps\)",
                                           img_format)[0]
    except ValueError:
      format_name, img_size, fps = img_format, None, None

    # Default color is BGR in most cases
    color = 'BGR'
    # Adding the decoder to the pipeline if needed
    if format_name == 'MJPG':
      img_format = '! jpegdec'
    elif format_name == 'H264':
      img_format = '! h264parse ! avdec_h264'
    elif format_name == 'HEVC':
      img_format = '! h265parse ! avdec_h265'
    elif format_name in ('YUYV', 'YUY2'):
      img_format = ''
    elif format_name == 'GRAY8':
      img_format = ''
      color = 'GRAY8'
    else:
      self.log(logging.WARNING, f"Unsupported format name: {format_name}, "
                                f"trying without explicitly setting the "
                                f"decoder in the pipeline")
      img_format = ''

    # Getting the width and height from the second half of the string
    if img_size is not None:
      width, height = map(int, img_size.split('x'))
    else:
      width, height = None, None

    # Including the dimensions and the fps in the pipeline
    img_size = f',width={width},height={height}' if width else ''
    if fps is not None:
      fps = Fraction(fps)
      fps_str = f',framerate={fps.numerator}/{fps.denominator}'
    else:
      fps_str = ''
    # Finally, generate a single pipeline containing all the user settings
    return f"""{source} {device} name=source {img_format} ! videoconvert ! 
           video/x-raw,format={color}{img_size}{fps_str} ! videobalance 
           brightness={brightness if brightness is not None 
                       else self.brightness:.3f} 
           contrast={contrast if contrast is not None else self.contrast:.3f} 
           hue={hue if hue is not None else self.hue:.3f}
           saturation={saturation if saturation is not None 
                       else self.saturation:.3f} ! 
           appsink name=sink"""

  def _on_new_sample(self, app_sink):
    """Callback that reads every new frame and puts it into a buffer.

    Args:
      app_sink: The AppSink object containing the new frames.

    Returns:
      A GStreamer object indicating that the reading went fine.
    """

    sample = app_sink.pull_sample()
    caps = sample.get_caps()

    # Extracting the width and height info from the sample's caps
    height = caps.get_structure(0).get_value("height")
    width = caps.get_structure(0).get_value("width")

    # Getting the actual data
    buffer = sample.get_buffer()
    # Getting read access to the buffer data
    success, map_info = buffer.map(Gst.MapFlags.READ)
    if not success:
      raise RuntimeError("Could not map buffer data!")
    self.log(logging.DEBUG, "Grabbed new frame")

    # Casting the data into a numpy array
    try:
      numpy_frame = ndarray(
        shape=(height, width, self._nb_channels),
        dtype=uint8 if self._img_depth == 8 else uint16,
        buffer=map_info.data)
    except TypeError:
      raise TypeError("Unexpected number of channels in the received image !\n"
                      "You can try adding something like ' ! videoconvert ! "
                      "video/x-raw,format=BGR ! ' before your sink to specify "
                      "the format.\n(here BGR would be for 3 channels)")

    # Converting to gray level if needed
    if (self._user_pipeline is None
        and hasattr(self, 'channels')
        and self.channels == '1'):
      numpy_frame = cv2.cvtColor(numpy_frame, cv2.COLOR_BGR2GRAY)

    # Cleaning up the buffer mapping
    buffer.unmap(map_info)

    self._img = squeeze(numpy_frame)
    self._frame_nr += 1

    return Gst.FlowReturn.OK

  def _set_brightness(self, brightness: float) -> None:
    """Sets the image brightness."""

    self._restart_pipeline(self._get_pipeline(brightness=brightness))

  def _set_contrast(self, contrast: float) -> None:
    """Sets the image contrast."""

    self._restart_pipeline(self._get_pipeline(contrast=contrast))

  def _set_hue(self, hue: float) -> None:
    """Sets the image hue."""

    self._restart_pipeline(self._get_pipeline(hue=hue))

  def _set_saturation(self, saturation: float) -> None:
    """Sets the image saturation."""

    self._restart_pipeline(self._get_pipeline(saturation=saturation))

  def _set_format(self, img_format) -> None:
    """Sets the image encoding and dimensions."""

    self._restart_pipeline(self._get_pipeline(img_format=img_format))

    # Reloading the software ROI selection settings
    if self._soft_roi_set and self._formats and ' ' in self._formats[0]:
      width, height = search(r'(\d+)x(\d+)', img_format).groups()
      self.reload_software_roi(int(width), int(height))



# coding: utf-8

from time import time
from typing import List
import logging
from  warnings import warn

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


class Agilent34420a(InOut):
  """This class can read resistance and voltage values from an Agilent 34420A
  multimeter.

  It communicates over serial.

  May also work on similar devices from the same manufacturer, although that
  was not tested.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               mode: bytes = b"VOLT",
               device: str = '/dev/ttyUSB0',
               baudrate: int = 9600,
               timeout: float = 1) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      mode: Measurement mode, as :obj:`bytes`. Should be either `b'VOLT'` or
        `b'RES'`.
      device: Path to the serial port to open, as a :obj:`str`.
      baudrate: Desired baudrate for serial communication.
      timeout: Timeout for the serial connection, as a :obj:`float`.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._ser = None

    super().__init__()

    self._device = device
    self._baudrate = baudrate
    self._timeout = timeout
    self._mode = mode

  def open(self) -> None:
    """Opens the serial connection, resets the Agilent and configures it to the
    desired mode."""

    self.log(logging.INFO, f"Opening the serial port {self._device} with "
                           f"baudrate {self._baudrate}")
    self._ser = serial.Serial(port=self._device, baudrate=self._baudrate,
                              timeout=self._timeout)
    self.log(logging.DEBUG, f"Writing b'*RST;*CLS;*OPC?\\n' to port "
                            f"{self._device}")
    self._ser.write(b"*RST;*CLS;*OPC?\n")
    self.log(logging.DEBUG, f"Writing b'SENS:FUNC \"{self._mode}\";  \\n' "
                            f"to port {self._device}")
    self._ser.write(b"SENS:FUNC \"" + self._mode + b"\";  \n")
    self.log(logging.DEBUG, f"Writing b'SENS:{self._mode}:NPLC 2  \\n' "
                            f"to port {self._device}")
    self._ser.write(b"SENS:" + self._mode + b":NPLC 2  \n")
    self.log(logging.DEBUG, f"Writing b'SYST:REM\\n' to port {self._device}")
    self._ser.write(b"SYST:REM\n")

  def get_data(self) -> List[float]:
    """Asks the Agilent to acquire a value and returns it, except if an error
    occurs in which case `0` is returned."""

    self.log(logging.DEBUG, f"Writing b'READ?  \\n' to port {self._device}")
    self._ser.write(b"READ?  \n")
    t = time()
    try:
      return [t, float(self._ser.readline())]
    except (serial.SerialException, ValueError):
      self._ser.flush()
      return [t, 0]

  def close(self) -> None:
    """Closes the serial port."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._device}")
      self._ser.close()



# coding: utf-8

"""This module contains the code specific to applications at the LaMcube
laboratory.

They are included in Crappy as most users are from the LaMcube, but kept in a
separate submodule because they're almost certainly never going to be used
as-is by anyone outside the lab.
"""

from .biaxe import Biaxe
from .bispectral import BiSpectral
from .biotens import Biotens



# coding: utf-8

from .auto_drive_video_extenso import AutoDriveVideoExtenso
from .meta_block import Block
from .camera import Camera
from .client_server import ClientServer
from .dashboard import Dashboard
from .dis_correl import DISCorrel
from .dic_ve import DICVE
from .canvas import Canvas
from .fake_machine import FakeMachine
from .generator import Generator
from .gpu_correl import GPUCorrel
from .gpu_ve import GPUVE
from .grapher import Grapher
from .button import Button
from .hdf_recorder import HDFRecorder
from .ioblock import IOBlock
from .machine import Machine
from .mean import MeanBlock
from .multiplexer import Multiplexer
from .pid import PID
from .link_reader import LinkReader
from .recorder import Recorder
from .sink import Sink
from .stop_block import StopBlock
from .stop_button import StopButton
from .ucontroller import UController
from .video_extenso import VideoExtenso

from . import generator_path
from . import camera_processes

from ._deprecated import (AutoDrive, Client_server, Displayer, DISVE, Drawing,
                          Fake_machine, GUI, Hdf_recorder, Mean_block,
                          Multiplex, Reader, Video_extenso)



# coding: utf-8

from time import time, sleep
from typing import Dict, Any, Optional, Iterator, Iterable
from itertools import cycle
from copy import deepcopy
import logging

from .meta_block import Block
from .generator_path.meta_path import paths_dict, Path
from .._global import GeneratorStop


class GeneratorNoStop(Exception):
  """A custom exception for handling the case when the Generator should not
  raise a :exc:`CrappyStop` exception when it terminates."""


class Generator(Block):
  """This Block generates a signal following a user-defined assembly of
  :class:`~crappy.blocks.generator_path.meta_path.Path`.
  
  The generated signal is just a waveform that can serve any purpose. It can
  for example be used for driving a :class:`~crappy.blocks.Machine` Block, or 
  for triggering a :class:`~crappy.blocks.Camera` Block.

  One Generator Block can only generate one signal. Use multiple Blocks if
  several signals are needed. Note that the default behavior of a Generator
  is to stop the entire script when it reaches the end of all the Paths.

  This Block can also accept inputs from other Blocks, as these inputs may be
  used by a :class:`~crappy.blocks.generator_path.meta_path.Path`. The most
  common use of this feature is to have the stop condition of a Path depend on
  the received values of a label.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               path: Iterable[Dict[str, Any]],
               freq: Optional[float] = 200,
               cmd_label: str = 'cmd',
               path_index_label: str = 'index',
               repeat: bool = False,
               spam: bool = False,
               display_freq: bool = False,
               end_delay: Optional[float] = 2,
               safe_start: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      path: An iterable (like a :obj:`list` or a :obj:`tuple`) of :obj:`dict`,
        each dict providing the parameters to generate a 
        :class:`~crappy.blocks.generator_path.meta_path.Path`. The Paths are
        generated in the order in which they are given, and the stop condition
        of each Path is used for determining when to switch to the next one.
        The ``'type'`` key of each :obj:`dict` gives the name of the Path to
        use, and all the other keys correspond to the arguments to give to
        this Path. Refer to the documentation of the chosen Paths to know which
        keys to provide.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      cmd_label: The label of the signal sent to the downstream Blocks.
      path_index_label: In addition to the ``cmd_label``, this label holds the
        index of the current
        :class:`~crappy.blocks.generator_path.meta_path.Path`. Useful to
        trigger a Block when the current Path changes, as the output value
        might not necessarily change.
      repeat: If :obj:`True`, the ``path`` will loop forever instead of
        stopping when it reaches the last Path.
      spam: If :obj:`True`, the signal value will be sent on each loop. Else,
        it will only be sent if it is different from the previous or if the
        Block switched to the next Path.
      display_freq: if :obj:`True`, displays the looping frequency of the 
        Block.
      end_delay: When all the Paths are exhausted, waits this many seconds
        before stopping the entire script. Can be set to :obj:`None`,
        in which case the Generator won't stop the program when finishing.
      safe_start: Ensures the first Path waits for at least one data point from
        upstream Blocks before sending the first value of the signal.
        Otherwise, the first value might be sent without checking the
        associated condition if its depends on labels from other Blocks.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionremoved:: 1.5.10 *cmd* and *trig_link* argument
    .. versionadded:: 1.5.10 *safe_start* argument
    .. versionadded:: 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()

    # Instantiating a few attributes based on the arguments
    self.niceness = -5
    self.freq = freq
    self.display_freq = display_freq
    self.labels = ['t(s)', cmd_label, path_index_label]
    self.debug = debug

    self._end_delay = end_delay
    self._spam = spam
    self._safe_start = safe_start
    self._safe_started = False

    # The path is an iterable object
    path = list(path)
    self._path = cycle(path) if repeat else iter(path)

    # More attributes
    self._ended_no_raise = False
    self._last_cmd = None
    self._last_id = None
    self._last_t = None
    self._current_path = None
    self._path_id = None

    # Checking the validity of the path
    self._check_path_validity(iter(deepcopy(iter(path))))

  def begin(self) -> None:
    """Initializes the first
    :class:`~crappy.blocks.generator_path.meta_path.Path`."""

    self._update_path()

  def loop(self) -> None:
    """First reads data from upstream Blocks, then gets the next command to
    send, and finally sends it to downstream Blocks.

    It also manages the transitions between the
    :class:`~crappy.blocks.generator_path.meta_path.Path`.
    """

    # Case when the Generator shouldn't raise CrappyStop after it ended
    if self._ended_no_raise:
      return

    # If self start requested, do nothing until the first values are received
    if self._safe_start and not self._safe_started:
      if self.data_available():
        self._safe_started = True
        self.log(logging.INFO, "First data received, starting safely")
      else:
        self.log(logging.DEBUG, "Waiting for first data to arrive for "
                                "starting safely")
        return

    # Getting the data from upstream blocks
    data = self.recv_all_data()
    try:
      # Getting the next command to send
      self._last_t = time()
      cmd = self._current_path.get_cmd(data)
      self.log(logging.DEBUG, f"Returned command: {cmd}")
    except StopIteration:
      try:
        # Switching to the next path if we reached the end of one
        self._update_path()
        self.loop()
        return
      except GeneratorNoStop:
        self.log(logging.WARNING, f"Generator path exhausted, staying idle "
                                  f"until the script ends")
        # Case when the Generator shouldn't raise CrappyStop after it ended
        self._ended_no_raise = True
        return

    # Not sending if no command was output by the get_cmd method
    if cmd is None:
      return

    # The command is sent if it's different from the previous, or if we
    # switched to the next path, or if spam is True
    if cmd != self._last_cmd or self._last_id != self._path_id or self._spam:
      self._last_cmd = cmd
      self._last_id = self._path_id
      # Actually sending the command
      self.send([self._last_t - self.t0, cmd, self._path_id])

  def _update_path(self) -> None:
    """Gets the next Path from the list of Paths and instantiates it.

    Also manages the case when the last Path of the list was reached.
    """

    try:
      # Getting the next path from the list of paths
      next_path_dict = deepcopy(next(self._path))
      # Updating the path index
      if self._path_id is not None:
        self._path_id += 1
      else:
        self._path_id = 0

    # Raised when the list of paths is exhausted
    except StopIteration:
      # First option, stopping the program after a delay
      if self._end_delay is not None:
        sleep(self._end_delay)
        raise GeneratorStop
      # Second option, not stopping the program and looping forever
      else:
        raise GeneratorNoStop

    self.log(logging.INFO, f"Next generator path (id: {self._path_id}): "
                           f"{next_path_dict['type']}")

    # Instantiating the next generator path object
    path_name = next_path_dict.pop('type')
    self._check_path_exists(path_name)
    path_type = paths_dict[path_name]
    Path.t0 = self._last_t if self._last_t is not None else self.t0
    Path.last_cmd = self._last_cmd
    self._current_path = path_type(**next_path_dict)

  def _check_path_validity(self, path: Iterator[Dict[str, Any]]) -> None:
    """Simply instantiates all the Paths in a row to check no error is
    raised."""

    for i, next_dict in enumerate(path):
      next_dict = deepcopy(next_dict)
      path_name = next_dict.pop('type')
      self._check_path_exists(path_name)
      Path.t0 = 0
      Path.last_cmd = None if i == 0 else 0
      path_type = paths_dict[path_name]
      path_type(**next_dict)

  @staticmethod
  def _check_path_exists(name: str) -> None:
    """Checks that the provided Generator Path is a valid one, and raises an
    error if not."""

    if name not in paths_dict:
      possible = ', '.join(sorted(paths_dict.keys()))
      raise ValueError(f"Unknown Generator path type : {name} ! "
                       f"The possible types are : {possible}")



# coding: utf-8

from typing import Optional

from .meta_block import Block


class Sink(Block):
  """This Block drops all the data it receives, and does nothing else.
  
  It is only useful for debugging, e.g. with Blocks like the
  :class:`~crappy.blocks.IOBlock` that have a different behavior when they have
  output Links.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self,
               display_freq: bool = False,
               freq: Optional[float] = 10,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.
    
    Args:
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    
    .. versionadded 1.5.6 *verbose* and *freq* arguments
    .. versionadded 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()
    self.display_freq = display_freq
    self.freq = freq
    self.debug = debug

  def loop(self) -> None:
    """Drops all the received data."""

    self.recv_all_data_raw()



# coding: utf-8

from typing import Callable, Optional, Union, Any
from multiprocessing import current_process
import logging

NbrType = Union[int, float]


class CameraSetting:
  """Base class for each Camera setting.

  It is meant to be subclassed and should not be used as is.

  The Camera setting classes hold all the information needed to read and set a
  setting of a :class:`~crappy.camera.Camera` object. Several types of settings
  are defined, as children of this class :
  :class:`~crappy.camera.meta_camera.camera_setting.CameraBoolSetting`,
  :class:`~crappy.camera.meta_camera.camera_setting.CameraChoiceSetting`,
  and :class:`~crappy.camera.meta_camera.camera_setting.CameraScaleSetting`.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Camera_setting to CameraSetting
  """

  def __init__(self,
               name: str,
               getter: Callable[[], Any],
               setter: Callable[[Any], None],
               default: Any) -> None:
    """Sets the attributes.

    Args:
      name: The name of the setting, that will be displayed in the GUI.
      getter: The method for getting the current value of the setting.
      setter: The method for setting the current value of the setting.
      default: The default value to assign to the setting.
    """

    # Attributes shared by all the settings
    self.name = name
    self.default = default
    self.type = type(default)

    # Attributes used in the GUI
    self.tk_var = None
    self.tk_obj = None

    # Attributes for internal use only
    self._value_no_getter = default
    self._getter = getter
    self._setter = setter
    self._logger: Optional[logging.Logger] = None

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the CameraSetting.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.

    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  @property
  def value(self) -> Any:
    """Returns the current value of the setting, by calling the getter if one
    was provided or else by returning the stored value.

    When the getter is called, calls the setter if one was provided and updates
    the sored value. After calling the setter, checks that the value was set
    by calling the getter and displays a warning message if the target and
    actual values don't match.
    """

    if self._getter is not None:
      return self._getter()
    else:
      return self._value_no_getter

  @value.setter
  def value(self, val: Any) -> None:
    self.log(logging.DEBUG, f"Setting the setting {self.name} to {val}")
    self._value_no_getter = val
    if self._setter is not None:
      self._setter(val)

    if self.value != val:
      # Double-checking, got strange behavior sometimes probably because of
      # delays in lower level APIs
      if self.value == val:
        return
      self.log(logging.WARNING, f"Could not set {self.name} to {val}, the "
                                f"value is {self.value} !")

  def reload(self, *_, **__) -> None:
    """Allows modifying a setting once it is already being displayed in the
    GUI.

    Mostly helpful for adjusting the ranges of sliders.

    .. versionadded:: 2.0.0
    """

    ...



# coding: utf-8

from time import time, sleep
from typing import Optional, Dict, Any, Union, List, Iterable
import numpy as np
import logging
from multiprocessing import current_process
from collections import defaultdict

from .meta_inout import MetaIO


class InOut(metaclass=MetaIO):
  """Base class for all InOut objects. Implements methods shared by all the
  InOuts, and ensures their dataclass is MetaIO.

  The InOut objects are helper classes used by the
  :class:`~crappy.blocks.IOBlock` to interface with hardware.
  
  .. versionadded:: 1.4.0
  """

  ft232h: bool = False

  def __init__(self, *_, **__) -> None:
    """Sets the attributes.
    
    .. versionchanged:: 2.0.0 now accepts args and kwargs
    """

    self._compensations: List[float] = list()
    self._compensations_dict: Dict[str, float] = dict()
    self._logger: Optional[logging.Logger] = None

  def log(self, level: int, msg: str) -> None:
    """Records log messages for the InOut.

    Also instantiates the logger when logging the first message.

    Args:
      level: An :obj:`int` indicating the logging level of the message.
      msg: The message to log, as a :obj:`str`.
    
    .. versionadded:: 2.0.0
    """

    if self._logger is None:
      self._logger = logging.getLogger(
        f"{current_process().name}.{type(self).__name__}")

    self._logger.log(level, msg)

  def open(self) -> None:
    """This method should perform any action that's required for initializing
    the hardware and the communication with it.

    Communication with hardware should be avoided in the :meth:`__init__`
    method, and this method is where it should start. This method is called
    after Crappy's processes start, i.e. when the associated
    :class:`~crappy.blocks.IOBlock` already runs separately from all the other
    Blocks.

    It is fine for this method not to perform anything.

    .. versionadded:: 1.5.10
    """

    ...

  def get_data(self) -> Optional[Union[Iterable, Dict[str, Any]]]:
    """This method should acquire data from a device and return it along with a
    timestamp.

    The data can be either returned as raw values in an iterable object (like
    a :obj:`list` or a :obj:`tuple`), or along with labels in a :obj:`dict`.
    Values can be of any type (:obj:`float`, :obj:`int`, :obj:`str`,
    :obj:`bytes`, etc.), but most Blocks in Crappy will expect numeric values.

    If the data is returned in a dictionary, its keys will be the labels for
    sending data to downstream Blocks. The ``labels`` argument of the
    :class:`~crappy.blocks.IOBlock` is ignored. It is **mandatory** to use
    `'t(s)'` as the label for the time value.

    Example:
      ::

        return {'t(s)': time.time(), 'value_1': -3.5, 'value_2': 4.1}

    If the data is returned as another type of iterable, the labels must be
    provided in the ``labels`` argument of the :class:`~crappy.blocks.IOBlock`.
    The number of labels must match the number of returned values. It is
    **mandatory** to return the time value first in the iterable.

    Example:
      ::

        return time.time(), -3.5, 4.1

    In both cases, there's no limit to the number of returned values. The same
    number of values must always be returned, and each value must keep a
    consistent type throughout the test.

    It is alright for this method to return :obj:`None` if there's no data to
    acquire.
    
    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING,
             "The get_data method was called but is not defined !\n To get "
             "rid of this warning, define a get_data method, or remove all "
             "the downstream links and make sure the make_zero_delay argument "
             "of the IOBlock is set to None")
    sleep(1)
    return

  def set_cmd(self, *cmd) -> None:
    """This method should handle commands received from the upstream Blocks.

    Among all the data received on the incoming Link(s), only the labels listed
    in the ``cmd_labels`` argument of the :class:`~crappy.blocks.IOBlock` are
    considered. The command values are passed as arguments to this method in
    the same order as the ``cmd_labels`` are given. They are passed as
    positional arguments, not keyword arguments.

    Example:
      If ``{'t(s)': 20.5, 'value_1': 1.5, 'value_2': 3.6, 'value_3': 'OK'}`` is
      received from upstream Blocks, and ``cmd_labels=('value_3', 'value_1')``,
      then the call will be ``set_cmd('OK', 1.5)``.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING,
             "The set_cmd method was called but is not defined ! The data "
             "received from the incoming links is discarded")
    sleep(1)
    return

  def start_stream(self) -> None:
    """This method should start the acquisition of the stream.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, "The start_stream method was called but is not "
                              "defined !")

  def get_stream(self) -> Optional[Union[Iterable[np.ndarray],
                                         Dict[str, np.ndarray]]]:
    """This method should acquire a stream as a :obj:`numpy.array`, and return
    it along with another array carrying the timestamps.

    Two arrays should be created: one of dimension `(m,)` carrying the
    timestamps, and one of dimension `(m, n)` carrying the acquired data. They
    represent `n` channels acquired at `m` successive time points. There's no
    maximum number of acquired channels, but this number must be constant
    throughout a test.

    The two arrays can either be returned as is in an iterable object (like a
    :obj:`list` or a :obj:`tuple`), or along with labels in a :obj:`dict`.

    If the arrays are returned in a dictionary, its keys will be the labels for
    sending data to downstream Blocks. The ``labels`` argument of the
    :class:`~crappy.blocks.IOBlock` is ignored. It is **mandatory** to use
    `'t(s)'` as the label for the time array.

    If the arrays are returned in another type of iterable, two labels must be
    provided in the ``labels`` argument of the :class:`~crappy.blocks.IOBlock`.
    The first label **must be** `'t(s)'`, the second can be anything. It is
    **mandatory** to return the time value first in the iterable.

    It is alright for this method to return :obj:`None` if there's no data to
    acquire.

    Note:
      It is technically  possible to return more than two arrays, or even
      objects that are not arrays, but it is not the intended use for this
      method and might interfere with some functionalities of Crappy.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, "The get_stream method was called but is not "
                              "defined ! No data sent to downstream links")
    sleep(1)
    return

  def stop_stream(self) -> None:
    """This method should stop the acquisition of the stream.

    .. versionadded:: 1.5.10
    """

    self.log(logging.WARNING, "The stop_stream method was called but is not "
                              "defined !")

  def close(self) -> None:
    """This method should perform any action required for properly ending the
    test and closing the communication with hardware.

    It will be called when the associated :class:`~crappy.blocks.IOBlock`
    receives the order to stop (usually because the user hit `CTRL+C`, or
    because a :class:`~crappy.blocks.Generator` Block reached the end of its
    path, or because an exception was raised in any of the Blocks).

    It is fine for this method not to perform anything.

    .. versionadded:: 1.5.10
    """

    ...

  def make_zero(self, delay: float) -> None:
    """This method acquires data for a given delay and stores for each channel
    the average value.

    These values are used for zeroing the channels, so that their values start
    at `0` in the beginning of the test.

    Important:
      This method uses :meth:`get_data` for acquiring the values, so it doesn't
      work for pure streams. It also doesn't work if the acquired values do not
      support arithmetic operations (like :obj:`str`).

    .. versionadded:: 1.5.10
    """

    buf = []
    buf_dict = defaultdict(list)
    t0 = time()

    # Acquiring data for a given delay
    while time() < t0 + delay:
      data = self.get_data()
      if data is not None:

        # Case when the InOut returns a dict
        if isinstance(data, dict):
          for label, value in data.items():
            buf_dict[label].append(value)

        # Case when the InOut returns another type of iterable
        else:
          data = list(data)
          if len(data) > 1:
            buf.append(data[1:])

    # Removing the time information if given
    if 't(s)' in buf_dict:
      del buf_dict['t(s)']

    # If no data could be acquired, abort
    if not buf and not buf_dict:
      self.log(logging.WARNING, "No data acquired when zeroing the channels, "
                                "aborting the zeroing")
      return

    # Averaging the values and storing them in a list
    if buf:
      for values in zip(*buf):
        try:
          self._compensations.append(-sum(values) / len(values))
        except TypeError:
          # If something goes wrong, just forget about the offsetting
          self._compensations = list()
          self.log(logging.WARNING,
                   "Cannot calculate the offset !\nA possible reason is that "
                   "the InOut doesn't return only numbers")
          return

    # Averaging the values and storing them in a dict
    else:
      for label, values in buf_dict.items():
        try:
          self._compensations_dict[label] = -sum(values) / len(values)
        except TypeError:
          # If something goes wrong, just forget about the offsetting
          self._compensations_dict = dict()
          self.log(logging.WARNING,
                   "Cannot calculate the offset !\nA possible reason is that "
                   "the InOut doesn't return only numbers in the dict")
          return

  def return_data(self) -> Optional[Union[List[Any], Dict[str, Any]]]:
    """Returns the data from :meth:`get_data`, corrected by an offset if the
    ``make_zero_delay`` argument of the :class:`~crappy.blocks.IOBlock` is
    set.

    .. versionadded:: 1.5.10
    """

    data = self.get_data()

    # Nothing to do if there's no data
    if data is None:
      return

    # Case when the data is returned as a dict
    if isinstance(data, dict):
      # Checking that the time label is given
      if 't(s)' not in data:
        raise ValueError("The time label 't(s)' must be given in the dict "
                         "returned by get_data() !")

      # If there's no offsetting, just return the data
      if not self._compensations_dict:
        return data

      # Offsetting if all the labels in data have an offset, except time label
      elif all(label in self._compensations_dict
               for label in data if label != 't(s)'):
        t = {'t(s)': data['t(s)']}
        comp = {label: data[label] + self._compensations_dict[label]
                for label in data if label != 't(s)'}
        return dict(**t, **comp)

      else:
        raise ValueError("Not all the labels in the acquired data have an "
                         "offset value !")

    # Case when data is returned as an iterable but not a dict
    else:
      # Converting to list for convenience
      data = list(data)

      # If there's no offsetting, just return the data
      if not self._compensations:
        return data

      # Offsetting if the number of acquired values match the number of offsets
      elif len(data[1:]) == len(self._compensations):
        return [data[0]] + [val + comp for val, comp
                            in zip(data[1:], self._compensations)]

      else:
        raise ValueError("The number of offsets doesn't match the number of "
                         "acquired values.")

  def return_stream(self) -> Optional[Union[List[np.ndarray],
                                            Dict[str, np.ndarray]]]:
    """Returns the data from :meth:`get_stream`, corrected by an offset if the
    ``make_zero_delay`` argument of the :class:`~crappy.blocks.IOBlock` is
    set.

    .. versionadded:: 1.5.10
    """

    data = self.get_stream()

    # Nothing to do if there's no data
    if data is None:
      return

    # Case when the stream is returned as a dict
    if isinstance(data, dict):
      # Checking that the time label is given
      if 't(s)' not in data:
        raise ValueError("The time label 't(s)' must be given in the dict "
                         "returned by get_stream() !")

      # If there's no offsetting, just return the stream
      if not self._compensations and not self._compensations_dict:
        return data

      t = {'t(s)': data['t(s)']}
      comp = dict()
      # Offsetting if the shape of the acquired values match the number of
      # offsets
      for label, array in ((key, val) for key, val in data.items()
                           if key != 't(s)'):
        # Case when get_data returns an iterable
        if self._compensations and len(self._compensations) == array.shape[1]:
          comp[label] = array + self._compensations
        # Case when get_data returns a dict
        elif (self._compensations_dict and
              len(self._compensations_dict) == array.shape[1]):
          comp[label] = array + list(self._compensations_dict.values())
        # The shapes do not match
        else:
          raise ValueError("The number of offsets doesn't match the shape of "
                           "the acquired array !")
        return dict(**t, **comp)

    # Case when stream is returned as an iterable but not a dict
    else:
      # Converting to list for convenience
      data = list(data)

      # If there's no offsetting, just return the stream
      if not self._compensations and not self._compensations_dict:
        return data

      # Offsetting if the shape of the acquired values match the number of
      # offsets
      comp = list()
      for array in data[1:]:
        # Case when get_data returns an iterable
        if self._compensations and len(self._compensations) == array.shape[1]:
          comp.append(array + self._compensations)
        # Case when get_data returns a dict
        elif (self._compensations_dict and
              len(self._compensations_dict) == array.shape[1]):
          comp.append(array + self._compensations_dict.values())
        # The shapes do not match
        else:
          raise ValueError("The number of offsets doesn't match the shape "
                           "of the acquired array !")
      return [data[0]] + comp



# coding: utf-8

from typing import Dict, Any, Union, Iterable
import logging

from .meta_modifier import Modifier


class Offset(Modifier):
  """This Modifier offsets every value of the given labels by a constant. This
  constant is calculated so that for each label the first returned value is
  equal to a user-defined target.

  For example if for a given label the target is `6` and the first received
  value is `3`, the Modifier will add `3` to each value received over this
  label.

  This Modifier can be used for example when measuring a variable that should
  start at `0` but doesn't because of a sensor offset. It can also just be used
  to plot nicer figures. It is not very accurate as it is only based on a
  single data point for the offset calculation. The ``make_zero`` argument of
  the :class:`~crappy.blocks.IOBlock` is a better alternative if precision is
  required when offsetting a sensor.
  
  .. versionadded:: 1.5.10
  """

  def __init__(self,
               labels: Union[str, Iterable[str]],
               offsets: Union[float, Iterable[float]]) -> None:
    """Sets the args and initializes the parent class.

    Args:
      labels: The labels to offset. Can be given as a single label, a
        :obj:`list` of labels or a :obj:`tuple` of labels.
      offsets: For each label, the target for the first received value. Can be
        given as a single value, a :obj:`list` of values or a :obj:`tuple` of
        values.
    """

    super().__init__()

    # Handling the case when only one label needs to be offset
    if isinstance(labels, str):
      labels = (labels,)
    labels = tuple(labels)
    try:
      iter(offsets)
    except TypeError:
      offsets = (offsets,)
    offsets = tuple(offsets)

    # Checking that the number of offsets match the number of labels
    if len(offsets) != len(labels):
      raise ValueError("As many offsets as there are labels should be given.")

    # Associating each offset to its label
    self._offsets = {label: offset for label, offset in zip(labels, offsets)}

    self._compensations = None
    self._compensated = False

  def __call__(self, data: Dict[str, Any]) -> Dict[str, Any]:
    """If the compensations are not set, sets them, and then offsets the
    required labels.
    
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    # During the first loop, calculating the compensation values
    if not self._compensated:
      self._compensations = {label: -data[label] + offset for label, offset in
                             self._offsets.items()}
      self._compensated = True

    # Compensating the data to match the target offset value
    for label in self._offsets:
      data[label] += self._compensations[label]

    self.log(logging.DEBUG, f"Sending {data}")
    return data



# coding: utf-8

import numpy as np
from typing import Optional, List, Union
from pathlib import Path
import logging
import logging.handlers

from .camera_process import CameraProcess
from ...tool.image_processing import GPUCorrelTool


class GPUCorrelProcess(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can perform
  GPU-accelerated digital image correlation on a given mask of the acquired
  images, and calculate various fields from it.

  It is used by the :class:`~crappy.blocks.GPUCorrel` Block to parallelize the
  image processing and the image acquisition. It delegates most of the
  computation to the :class:`~crappy.tool.image_processing.GPUCorrelTool`. It
  is from this class that the output values are sent to the downstream Blocks.

  It is also this class that takes the decision to send or not the results to
  downstream Blocks based on the value of the calculated residuals, if this
  option is enabled by the user.

  .. versionadded:: 2.0.0
  """

  def __init__(self,
               discard_limit: float = 3,
               discard_ref: int = 5,
               calc_res: bool = False,
               img_ref: Optional[np.ndarray] = None,
               verbose: int = 0,
               levels: int = 5,
               resampling_factor: float = 2,
               kernel_file: Optional[Union[str, Path]] = None,
               iterations: int = 4,
               fields: Optional[List[str]] = None,
               mask: Optional[np.ndarray] = None,
               mul: float = 3) -> None:
    """Sets the arguments and initializes the parent class.
    
    Args:
      discard_limit: If ``calc_res`` is :obj:`True`, the result of the
        correlation is not sent to the downstream Blocks if the residuals for
        the current image are greater than ``discard_limit`` times the average
        residual for the last ``discard_ref`` images.
      discard_ref: If ``calc_res`` is :obj:`True`, the result of the
        correlation is not sent to the downstream Blocks if the residuals for
        the current image are greater than ``discard_limit`` times the average
        residual for the last ``discard_ref`` images.
      calc_res: If :obj:`True`, calculates the residuals after performing the
        correlation and returns the residuals along with the correlation data.
        By default, the residuals are returned under the label ``'res'``.
      img_ref: A reference image for the correlation, given as a 2D
        :obj:`numpy.array` with `dtype` :obj:`numpy.float32`. If not given, the
        first acquired frame will be used as the reference image instead. This
        argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      verbose: The verbose level as an integer, between `0` and `3`. At level
        `0` no information is displayed, and at level `3` so much information
        is displayed that it slows the code down. This argument is passed to
        the :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used
        in this class.
      levels: Number of levels of the pyramid. More levels may help converging
        on images with large strain, but may fail on images that don't contain
        low spatial frequency. Fewer levels mean that the program runs faster.
        This argument is passed to  the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      resampling_factor: The factor by which the resolution is divided between
        each stage of the pyramid. A low factor ensures coherence between the
        stages, but is more computationally intensive. A high factor allows
        reaching a finer detail level, but may lead to a coherence loss between
        the stages. This argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      kernel_file: The path to the file containing the kernels to use for the
        correlation. Can be a :obj:`pathlib.Path` object or a :obj:`str`. If
        not provided, the default :ref:`GPU Kernels` are used. This argument is
        passed to the :class:`~crappy.tool.image_processing.GPUCorrelTool` and
        not used in this class.
      iterations: The maximum number of iterations to run before returning the
        results. The results may be returned before if the residuals start
        increasing. This argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      fields: A :obj:`list` of :obj:`str` representing the base of fields on
        which the image will be projected during correlation. The possible
        fields are :
        ::

          'x', 'y', 'r', 'exx', 'eyy', 'exy', 'eyx', 'exy2', 'z'

        This argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      mask: The mask used for weighting the region of interest on the image. It
        is generally used to prevent unexpected behavior on the border of the
        image. This argument is passed to the
        :class:`~crappy.tool.image_processing.GPUCorrelTool` and not used in
        this class.
      mul: The scalar by which the direction will be multiplied before being
        added to the solution. If it's too high, the convergence will be fast
        but there's a risk to go past the solution and to diverge. If it's too
        low, the convergence will be slower and require more iterations. `3`
        was found to be an acceptable value in most cases, but it is
        recommended to tune this value for each application so that the
        convergence is neither too slow nor too fast. This argument is passed
        to the :class:`~crappy.tool.image_processing.GPUCorrelTool` and not
        used in this class.
    """

    super().__init__()

    # Arguments to pass to the GPUCorrelTool
    self._verbose = verbose
    self._levels = levels
    self._resampling_factor = resampling_factor
    self._kernel_file = kernel_file
    self._iterations = iterations
    self._fields = fields
    self._mask = mask
    self._mul = mul

    # Other attributes
    self._correl: Optional[GPUCorrelTool] = None
    self._img_ref = img_ref
    self._img0_set = img_ref is not None

    self._res_history = [np.inf]
    self._discard_limit = discard_limit
    self._discard_ref = discard_ref
    self._calc_res = calc_res

  def init(self) -> None:
    """Initializes the GPUCorrelTool, and set its reference image if a
    ``img_ref`` argument was provided."""

    # Instantiating the GPUCorrelTool
    self.log(logging.INFO, "Instantiating the GPUCorrel tool")
    self._correl = GPUCorrelTool(logger_name=self.name,
                                 context=None,
                                 verbose=self._verbose,
                                 levels=self._levels,
                                 resampling_factor=self._resampling_factor,
                                 kernel_file=self._kernel_file,
                                 iterations=self._iterations,
                                 fields=self._fields,
                                 ref_img=self._img_ref,
                                 mask=self._mask,
                                 mul=self._mul)

    # Setting the reference image if it was given as an argument
    if self._img_ref is not None:
      self.log(logging.INFO, "Initializing the GPUCorrel tool with the "
                             "given reference image")
      self._correl.set_img_size(self._img_ref.shape)
      self._correl.set_orig(self._img_ref.astype(np.float32))
      self.log(logging.INFO, "Preparing the GPUCorrel tool")
      self._correl.prepare()

  def loop(self) -> None:
    """This method grabs the latest frame and gives it for processing to the
    :class:`~crappy.tool.image_processing.GPUCorrelTool`. Then sends the result
    of the correlation to the downstream Blocks.

    If there's no new frame grabbed, doesn't do anything. On the first acquired
    frame, does not process it but initializes the GPUCorrelTool with it
    instead if no reference image was given as argument. If requested by the
    user, also calculates the residuals and checks that they are within the
    provided limit. Otherwise, just drops the calculated data.
    """

    # Setting the reference image with the first received frame if it was not
    # given as an argument
    if not self._img0_set:
      self.log(logging.INFO, "Setting the reference image")
      self._correl.set_img_size(self.img.shape)
      self._correl.set_orig(self.img.astype(np.float32))
      self._correl.prepare()
      self._img0_set = True
      return

    # Performing the image correlation
    self.log(logging.DEBUG, "Processing the received image")
    data = [self.metadata['t(s)'], self.metadata]
    data += self._correl.get_disp(self.img.astype(np.float32)).tolist()

    # Calculating the residuals if requested
    if self._calc_res:
      self.log(logging.DEBUG, "Calculating the residuals")
      res = self._correl.get_res()
      data.append(res)

      # Checking that the residuals are within the given limit, otherwise
      # dropping the calculated data
      if self._discard_limit:
        self.log(logging.DEBUG, "Adding residuals to the residuals "
                                "history")
        self._res_history.append(res)
        self._res_history = self._res_history[-self._discard_ref - 1:]

        if res > self._discard_limit * np.average(self._res_history[:-1]):
          self.log(logging.WARNING, "Residual too high, not sending "
                                    "values")
          return

    # Sending the data to downstream Blocks
    self.send(data)

  def finish(self) -> None:
    """Performs cleanup on the
    :class:`~crappy.tool.image_processing.GPUCorrelTool`."""

    if self._correl is not None:
      self.log(logging.INFO, "Cleaning up the GPUCorrel tool")
      self._correl.clean()



# coding: utf-8

from time import time

from .meta_actuator import Actuator


class FakeDCMotor(Actuator):
  """Emulates the behavior of a DC electric machine, driven through its input
  voltage.

  It is mainly intended for testing scripts without requiring any hardware.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 Renamed from Fake_motor to FakeDCMotor
  """

  def __init__(self,
               inertia: float = 0.5,
               torque: float = 0.,
               kv: float = 1000,
               rv: float = 0.4,
               fv: float = 2e-5,
               simulation_speed: float = 1,
               initial_speed: float = 0,
               initial_pos: float = 0) -> None:
    """Sets the instance attributes.

    Args:
      inertia: The inertia of the motor, in `kg.m²`.
      torque: A constant torque applied on the shaft of the motor in `N.m`.
      kv: The electrical constant of the motor, in`t/min/V`.
      rv: The internal solid friction coefficient of the motor, no unit.
      fv: The internal fluid friction coefficient of the motor, no unit.
      simulation_speed: Speed factor of the simulation, to speed it up or slow
        it down.
      initial_speed: The initial speed of the motor, in RPM.
      initial_pos: The initial position of the motor, in turns.
    
    .. versionchanged:: 2.0.0
       renamed *sim_speed* argument to *simulation_speed*
    """

    super().__init__()

    self._inertia = inertia
    self._torque = torque
    self._kv = kv
    self._rv = rv
    self._fv = fv
    self._simulation_speed = simulation_speed
    self._initial_speed = initial_speed
    self._initial_pos = initial_pos

    self._rpm = 0
    self._pos = 0
    self._volt = 0
    self._t = time()

  def open(self) -> None:
    """Sets the variables describing the state of the motor."""

    self._rpm = self._initial_speed
    self._pos = self._initial_pos
    self._volt = 0
    self._t = time() * self._simulation_speed

  def get_speed(self) -> float:
    """Return the speed of the motor, in RPM."""

    self._update()
    return self._rpm

  def get_position(self) -> float:
    """Returns the position of the motor, in rounds.

    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    self._update()
    return self._pos

  def set_speed(self, volt: float) -> None:
    """Sets the command of the motor, in volts."""

    self._update()
    self._volt = volt

  def _update(self) -> None:
    """Updates the motor variables based on the timestamp and their previous
    values.

    It supposes that the voltage has been constant since the last update.
    """

    t1 = time() * self._simulation_speed
    dt = (t1 - self._t)
    self._t = t1

    f = self._volt * self._kv - self._torque - self._rpm * \
        (1 + self._rv + self._rpm * self._fv)
    drpm = f / self._inertia * dt
    self._pos += dt * (self._rpm + drpm / 2)
    self._rpm += drpm



# coding: utf-8

from .ads1115 import ADS1115FT232H
from .gpio_switch import GPIOSwitchFT232H
from .mcp9600 import MCP9600FT232H
from .mprls import MPRLSFT232H
from .nau7802 import NAU7802FT232H
from .waveshare_ad_da import WaveshareADDAFT232H



# coding: utf-8

from typing import Optional, Iterable, Union, Callable, Dict
import logging
from re import split
from time import time

from .meta_block import Block


class StopBlock(Block):
  """This Block parses the data it receives and checks if this data meets the
  given stop criteria. If so, its stops the test.

  Along with the :class:`~crappy.blocks.StopButton` Block, it allows to stop a
  test in a clean way without resorting to CTRL+C.

  .. versionadded 2.0.0
  """

  def __init__(self,
               criteria: Union[str, Callable, Iterable[Union[str, Callable]]],
               freq: Optional[float] = 30,
               display_freq: bool = False,
               debug: Optional[bool] = False
               ) -> None:
    """Sets the arguments and initialize the parent class.

    Args:
      criteria: A :obj:`str`, a :obj:`~typing.Callable`, or an
        :obj:`~typing.Iterable` (like a :obj:`tuple` or a :obj:`list`)
        containing such objects. Each :obj:`str` or :obj:`~typing.Callable`
        represents one stop criterion. There is no limit to the given number of
        stop criteria. If a criterion is given as an :obj:`~typing.Callable`,
        it should accept as its sole argument the output of the
        :meth:`crappy.blocks.Block.recv_all_data` method and return :obj:`True`
        if the criterion is met, and :obj:`False` otherwise. If the criterion
        is given as a :obj:`str`, it should follow the following syntax :
        ::

          '<lab> > <threshold>'
          '<lab> < <threshold>'

        With ``<lab>`` and ``<threshold>`` to be replaced respectively with the
        name of a received label, and a threshold value. The spaces in the
        string are ignored.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      display_freq: if :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    # Handling the case when only one stop condition is given
    if isinstance(criteria, str) or isinstance(criteria, Callable):
      criteria = (criteria,)
    criteria = tuple(criteria)

    self._raw_crit = criteria
    self._criteria = None

  def prepare(self) -> None:
    """Converts all the given criteria to :ref:`collections.abc.Callable`."""

    # This operation cannot be performed during __init__ due to limitations of
    # the spawn start method of multiprocessing
    self._criteria = tuple(map(self._parse_criterion, self._raw_crit))

  def loop(self) -> None:
    """Receives data from upstream Blocks, checks if this data meets the
    criteria, and stop the test if that's the case."""

    data = self.recv_all_data()

    if self._criteria and any(crit(data) for crit in self._criteria):
      self.log(logging.WARNING, "Stop criterion reached, stopping all the "
                                "Blocks !")
      self.stop()

    self.log(logging.DEBUG, "No stop criterion reached during this loop")
  
  def _parse_criterion(self, criterion: Union[str, Callable]) -> Callable:
    """Parses a Callable or string criterion given as an input by the user, and
    returns the associated Callable."""
    
    # If the criterion is already a callable, returning it
    if isinstance(criterion, Callable):
      self.log(logging.DEBUG, "Criterion is a callable")
      return criterion

    # Second case, the criterion is a string containing '<'
    if '<' in criterion:
      self.log(logging.DEBUG, "Criterion is of type var < thresh")
      var, thresh = split(r'\s*<\s*', criterion)

      # Return a function that checks if received data is inferior to threshold
      def cond(data: Dict[str, list]) -> bool:
        """Criterion checking that the label values are below a given
        threshold."""

        if var in data:
          return any((val < float(thresh) for val in data[var]))
        return False

      return cond

    # Third case, the criterion is a string containing '>'
    elif '>' in criterion:
      self.log(logging.DEBUG, "Criterion is of type var > thresh")
      var, thresh = split(r'\s*>\s*', criterion)

      # Special case for a time criterion
      if var == 't(s)':
        self.log(logging.DEBUG, "Criterion is about the elapsed time")

        # Return a function that checks if the given time was reached
        def cond(_: Dict[str, list]) -> bool:
          """Criterion checking if a given delay is expired."""

          return time() - self.t0 > float(thresh)

        return cond

      # Regular case
      else:

        # Return a function that checks if received data is superior to
        # threshold
        def cond(data: Dict[str, list]) -> bool:
          """Criterion checking that the label values are above a given
          threshold."""

          if var in data:
            return any((val > float(thresh) for val in data[var]))
          return False

        return cond

    # Otherwise, it's an invalid syntax
    else:
      raise ValueError("Wrong syntax for the criterion, please refer to the "
                       "documentation")
      



# coding: utf-8

import numpy as np
from typing import Tuple
import logging

from ..camera.cameralink import BaslerIronmanCameraLink

table = (0x0000, 0xC0C1, 0xC181, 0x0140, 0xC301, 0x03C0, 0x0280, 0xC241,
         0xC601, 0x06C0, 0x0780, 0xC741, 0x0500, 0xC5C1, 0xC481, 0x0440,
         0xCC01, 0x0CC0, 0x0D80, 0xCD41, 0x0F00, 0xCFC1, 0xCE81, 0x0E40,
         0x0A00, 0xCAC1, 0xCB81, 0x0B40, 0xC901, 0x09C0, 0x0880, 0xC841,
         0xD801, 0x18C0, 0x1980, 0xD941, 0x1B00, 0xDBC1, 0xDA81, 0x1A40,
         0x1E00, 0xDEC1, 0xDF81, 0x1F40, 0xDD01, 0x1DC0, 0x1C80, 0xDC41,
         0x1400, 0xD4C1, 0xD581, 0x1540, 0xD701, 0x17C0, 0x1680, 0xD641,
         0xD201, 0x12C0, 0x1380, 0xD341, 0x1100, 0xD1C1, 0xD081, 0x1040,
         0xF001, 0x30C0, 0x3180, 0xF141, 0x3300, 0xF3C1, 0xF281, 0x3240,
         0x3600, 0xF6C1, 0xF781, 0x3740, 0xF501, 0x35C0, 0x3480, 0xF441,
         0x3C00, 0xFCC1, 0xFD81, 0x3D40, 0xFF01, 0x3FC0, 0x3E80, 0xFE41,
         0xFA01, 0x3AC0, 0x3B80, 0xFB41, 0x3900, 0xF9C1, 0xF881, 0x3840,
         0x2800, 0xE8C1, 0xE981, 0x2940, 0xEB01, 0x2BC0, 0x2A80, 0xEA41,
         0xEE01, 0x2EC0, 0x2F80, 0xEF41, 0x2D00, 0xEDC1, 0xEC81, 0x2C40,
         0xE401, 0x24C0, 0x2580, 0xE541, 0x2700, 0xE7C1, 0xE681, 0x2640,
         0x2200, 0xE2C1, 0xE381, 0x2340, 0xE101, 0x21C0, 0x2080, 0xE041,
         0xA001, 0x60C0, 0x6180, 0xA141, 0x6300, 0xA3C1, 0xA281, 0x6240,
         0x6600, 0xA6C1, 0xA781, 0x6740, 0xA501, 0x65C0, 0x6480, 0xA441,
         0x6C00, 0xACC1, 0xAD81, 0x6D40, 0xAF01, 0x6FC0, 0x6E80, 0xAE41,
         0xAA01, 0x6AC0, 0x6B80, 0xAB41, 0x6900, 0xA9C1, 0xA881, 0x6840,
         0x7800, 0xB8C1, 0xB981, 0x7940, 0xBB01, 0x7BC0, 0x7A80, 0xBA41,
         0xBE01, 0x7EC0, 0x7F80, 0xBF41, 0x7D00, 0xBDC1, 0xBC81, 0x7C40,
         0xB401, 0x74C0, 0x7580, 0xB541, 0x7700, 0xB7C1, 0xB681, 0x7640,
         0x7200, 0xB2C1, 0xB381, 0x7340, 0xB101, 0x71C0, 0x7080, 0xB041,
         0x5000, 0x90C1, 0x9181, 0x5140, 0x9301, 0x53C0, 0x5280, 0x9241,
         0x9601, 0x56C0, 0x5780, 0x9741, 0x5500, 0x95C1, 0x9481, 0x5440,
         0x9C01, 0x5CC0, 0x5D80, 0x9D41, 0x5F00, 0x9FC1, 0x9E81, 0x5E40,
         0x5A00, 0x9AC1, 0x9B81, 0x5B40, 0x9901, 0x59C0, 0x5880, 0x9841,
         0x8801, 0x48C0, 0x4980, 0x8941, 0x4B00, 0x8BC1, 0x8A81, 0x4A40,
         0x4E00, 0x8EC1, 0x8F81, 0x4F40, 0x8D01, 0x4DC0, 0x4C80, 0x8C41,
         0x4400, 0x84C1, 0x8581, 0x4540, 0x8701, 0x47C0, 0x4680, 0x8641,
         0x8201, 0x42C0, 0x4380, 0x8341, 0x4100, 0x81C1, 0x8081, 0x4040)


def calc_string(st: str, crc: int) -> int:
    """Given a binary string and starting CRC, Calc a final CRC-16."""

    for ch in st:
        crc = (crc >> 8) ^ table[(crc ^ ord(ch)) & 0xFF]
    return crc


def add_crc(s: str) -> str:
  """Wrapper for adding a CRC to a serial command."""

  return s + hex(calc_string(s, 0xFFFF)).split('x')[1].upper().rjust(4, '0')


def check_crc(s: str) -> bool:
  """Wrapper for checking whether a CRC is valid."""

  r = s[:-4]
  return add_crc(r) == s


def hexlify(n: int) -> str:
  """Converts an integer to its hexadecimal representation."""

  return hex(n).split('x')[1].rjust(2, '0').upper()


class BiSpectral(BaslerIronmanCameraLink):
  """This class allows driving a bi-chromatic infrared camera, through a Basler
  microEnable 5 Ironman AD8 PoCL acquisition board.

  It is a child of the
  :class:`~crappy.camera.cameralink.BaslerIronmanCameraLink` Camera. It can set
  various settings on the camera, including the ROI or the trigger mode.

  The bi-chromatic camera is a very specific setup, and won't certainly be ever
  used outside the LaMcube laboratory. This is why it is stored in the LaMcube
  submodule.

  Warning:
    This Camera relies on a custom-written C library that hasn't been tested in
    a long time. It might not be functional anymore. This Camera also requires
    proprietary drivers to be installed.
    
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Bispectral to BiSpectral
  """

  def __init__(self) -> None:
    """Adds the various setting for the Camera."""

    super().__init__()
    self.add_scale_setting('width', 1, 640, self._get_w, self._set_w, 640)
    self.add_scale_setting('height', 1, 512, self._get_h, self._set_h, 512)
    self.add_scale_setting('xoffset', 0, 1278, self._get_ox, self._set_ox, 0)
    self.add_scale_setting('yoffset', 0, 511, self._get_oy, self._set_oy, 0)
    self.add_scale_setting('IT1', 10, 10000, self._get_it1, self._set_it1)
    self.add_scale_setting('IT2', 10, 10000, self._get_it2, self._set_it2)
    self.add_scale_setting('fps', 1., 150., self._get_trigg_freq,
                           self._set_trigg_freq)
    self.add_trigger_setting(setter=self._set_external_trigger)

  def open(self,
           camera_type: str = 'SingleAreaGray2DShading',
           **kwargs) -> None:
    """Opens the Camera and sends initialization commands.
    
    .. versionadded:: 1.5.10 explicitly listing *camera_type* argument
    """

    super().open(camera_type=camera_type, **kwargs)

    self._send_cmd('@W1A084')  # Restore unwindowed Mode
    self._send_cmd('@W10012')  # Make sure the image is not inverted

  def get_image(self) -> Tuple[float, np.ndarray]:
    """Grabs an image using the parent class' method, transforms it, and
    returns it."""

    t, frame = super().get_image()
    img = np.ones((self.height, self.width * 2), dtype=np.uint8)
    img[::, :self.width:2] = frame[::, ::4]
    img[::, 1:self.width:2] = frame[::, 1::4]
    img[::, self.width::2] = frame[::, 2::4]
    img[::, self.width + 1::2] = frame[::, 3::4]
    return t, img

  def _send_cmd(self, cmd: str) -> str:
    """Wrapper for sending a command to the Camera."""

    self.log(logging.DEBUG, f"Sending command {cmd}")
    r = self._cap.serialWrite(add_crc(cmd))
    if not check_crc(r) or r[1] != 'Y':
      self.log(logging.WARNING, f"Incorrect reply {r}")
    return r[2:4]

  def _set_external_trigger(self, val: str) -> None:
    """Sets the external trigger to val by toggling the value of the 3rd bit
    of register 102."""

    if val == 'Hardware':
      self._send_cmd('@W1027C')  # 3rd bit to 1
    else:
      self._send_cmd('@W10274')  # 3rd bit to 0

  def _get_roi(self) -> Tuple[int, int, int, int]:
    """Returns the minimum and maximum x and y coordinates of the current
    ROI."""

    x1min_lsb = self._send_cmd("@R1D0")
    x1min_msb = self._send_cmd("@R1D1")
    y1min_lsb = self._send_cmd("@R1D2")
    y1min_msb = self._send_cmd("@R1D3")
    x1max_lsb = self._send_cmd("@R1D4")
    x1max_msb = self._send_cmd("@R1D5")
    y1max_lsb = self._send_cmd("@R1D6")
    y1max_msb = self._send_cmd("@R1D7")
    xmin = int(x1min_msb + x1min_lsb, 16)
    xmax = int(x1max_msb + x1max_lsb, 16)
    ymin = int(y1min_msb + y1min_lsb, 16)
    ymax = int(y1max_msb + y1max_lsb, 16)
    return xmin, ymin, xmax, ymax

  def _set_roi(self, xmin: int, ymin: int, xmax: int, ymax: int) -> None:
    """Sets the minimum and maximum x and y coordinates of the ROI."""

    if (xmin, xmax, ymin, ymax) != (0, 0, 639, 511):
      self._send_cmd('@W1A080')  # Set to windowed mode
    else:
      self._send_cmd('@W1A084')
      self.log(logging.INFO, f"D set ROI to {xmin}, {ymin}, {xmax}, {ymax}")
    lsb_xmin = hexlify(xmin % 256)
    msb_xmin = hexlify(xmin // 256)
    lsb_xmax = hexlify(xmax % 256)
    msb_xmax = hexlify(xmax // 256)
    lsb_ymin = hexlify(ymin % 256)
    msb_ymin = hexlify(ymin // 256)
    lsb_ymax = hexlify(ymax % 256)
    msb_ymax = hexlify(ymax // 256)
    self._send_cmd("@W1D0" + lsb_xmin)
    self._send_cmd("@W1D1" + msb_xmin)
    self._send_cmd("@W1D2" + lsb_ymin)
    self._send_cmd("@W1D3" + msb_ymin)
    self._send_cmd("@W1D4" + lsb_xmax)
    self._send_cmd("@W1D5" + msb_xmax)
    self._send_cmd("@W1D6" + lsb_ymax)
    self._send_cmd("@W1D7" + msb_ymax)

  def _get_it(self) -> Tuple[float, float]:
    """Reads the integration time from the Camera and returns it."""

    mc = 10.35  # MHz
    it1_lsb = self._send_cmd("@R1B4")
    it1_mid = self._send_cmd("@R1B5")
    it1_msb = self._send_cmd("@R1B6")
    it2_lsb = self._send_cmd("@R1B8")
    it2_mid = self._send_cmd("@R1B9")
    it2_msb = self._send_cmd("@R1BA")
    it1 = int(it1_msb + it1_mid + it1_lsb, 16)  # Number of clock cycles
    it2 = int(it2_msb + it2_mid + it2_lsb, 16)
    return it1 / mc, it2 / mc  # IT in µs

  def _set_it(self, it1: int, it2: int) -> None:
    """Sets the integration time on the Camera."""

    mc = 10.35
    it1 = int(mc * it1)
    it2 = int(mc * it2)
    it1_lsb = hexlify(it1 % 256)
    it1 -= it1 % 256
    it1 //= 256
    it1_mid = hexlify(it1 % 256)
    it1_msb = hexlify(it1 // 256)
    it2_lsb = hexlify(it2 % 256)
    it2 -= it2 % 256
    it2 //= 256
    it2_mid = hexlify(it2 % 256)
    it2_msb = hexlify(it2 // 256)
    self._send_cmd("@W1B4" + it1_lsb)
    self._send_cmd("@W1B5" + it1_mid)
    self._send_cmd("@W1B6" + it1_msb)
    self._send_cmd("@W1B8" + it2_lsb)
    self._send_cmd("@W1B9" + it2_mid)
    self._send_cmd("@W1BA" + it2_msb)

  def _get_trigg_freq(self) -> float:
    """Reads the trigger frequency from the Camera."""

    mc = 10350000  # Hz
    p_lsb = self._send_cmd("@R1B0")
    p_mid = self._send_cmd("@R1B1")
    p_msb = self._send_cmd("@R1B2")
    p = int(p_msb + p_mid + p_lsb, 16)
    return mc / p

  def _set_trigg_freq(self, freq: float) -> None:
    """Sets the trigger frequency on the Camera."""

    mc = 10350000  # Hz
    period = int(mc / freq)
    p_lsb = hexlify(period % 256)
    period -= period % 256
    period //= 256
    p_mid = hexlify(period % 256)
    p_msb = hexlify(period // 256)
    self._send_cmd("@W1B0" + p_lsb)
    self._send_cmd("@W1B1" + p_mid)
    self._send_cmd("@W1B2" + p_msb)

  def _set_w(self, val: int) -> None:
    super()._set_w(val * 2)
    self._set_roi(self.xoffset, self.yoffset, self.xoffset + self.width - 1,
                  self.yoffset + self.height - 1)

  def _get_w(self) -> int:
    return int(super()._get_w() / 2)

  def _set_h(self, val: int) -> None:
    super()._set_h(val)
    self._set_roi(self.xoffset, self.yoffset, self.xoffset + self.width - 1,
                  self.yoffset + self.height - 1)

  def _set_ox(self, val: int) -> None:
    self._set_roi(val, self.yoffset, val + self.width - 1,
                  self.yoffset + self.height - 1)

  def _set_oy(self, val: int) -> None:
    self._set_roi(self.xoffset, val, self.xoffset + self.width - 1,
                  val + self.height - 1)

  def _get_ox(self) -> int:
    return self._get_roi()[0]

  def _get_oy(self) -> int:
    return self._get_roi()[1]

  def _get_it1(self) -> int:
    return int(self.get_itT()[0])

  def _get_it2(self) -> int:
    return int(self._get_it()[1])

  def _set_it1(self, val: int) -> None:
    self._set_it(val, self._get_it2())

  def _set_it2(self, val: int) -> None:
    self._set_it(self._get_it1(), val)



# coding: utf-8

from time import time, sleep
from typing import Optional, List
import logging

from .meta_inout import InOut
from .._global import OptionalModule

try:
  import smbus2
except (ModuleNotFoundError, ImportError):
  smbus2 = OptionalModule("smbus2")

try:
  import RPi.GPIO as GPIO
except (ModuleNotFoundError, ImportError):
  GPIO = OptionalModule("RPi.GPIO")

# Register Map
NAU7802_Scale_Registers = {'PU_CTRL': 0x00,
                           'CTRL1': 0x01,
                           'CTRL2': 0x02,
                           'OCAL1_B2': 0x03,
                           'OCAL1_B1': 0x04,
                           'OCAL1_B0': 0x05,
                           'GCAL1_B3': 0x06,
                           'GCAL1_B2': 0x07,
                           'GCAL1_B1': 0x08,
                           'GCAL1_B0': 0x09,
                           'OCAL2_B2': 0x0A,
                           'OCAL2_B1': 0x0B,
                           'OCAL2_B0': 0x0C,
                           'GCAL2_B3': 0x0D,
                           'GCAL2_B2': 0x0E,
                           'GCAL2_B1': 0x0F,
                           'GCAL2_B0': 0x10,
                           'I2C_CONTROL': 0x11,
                           'ADCO_B2': 0x12,
                           'ADCO_B1': 0x13,
                           'ADCO_B0': 0x14,
                           'ADC': 0x15,
                           'OTP_B1': 0x16,
                           'OTP_B0': 0x17,
                           'PGA': 0x1B,
                           'PGA_PWR': 0x1C,
                           'DEVICE_REV': 0x1F}

# Bits within the PU_CTRL register
NAU7802_PU_CTRL_Bits = {'PU_CTRL_RR': 0,
                        'PU_CTRL_PUD': 1,
                        'PU_CTRL_PUA': 2,
                        'PU_CTRL_PUR': 3,
                        'PU_CTRL_CS': 4,
                        'PU_CTRL_CR': 5,
                        'PU_CTRL_OSCS': 6,
                        'PU_CTRL_AVDDS': 7}

# Bits within the CTRL2 register
NAU7802_CTRL2_Bits = {'CTRL2_CALMOD': 0,
                      'CTRL2_CALS': 2,
                      'CTRL2_CAL_ERROR': 3,
                      'CTRL2_CRS': 4,
                      'CTRL2_CHS': 7}

# Bits within the PGA PWR register
NAU7802_PGA_PWR_Bits = {'PGA_PWR_PGA_CURR': 0,
                        'PGA_PWR_ADC_CURR': 2,
                        'PGA_PWR_MSTR_BIAS_CURR': 4,
                        'PGA_PWR_PGA_CAP_EN': 7}

# Allowed Low drop out regulator voltages
NAU7802_LDO_Values = {2.4: 0b111,
                      2.7: 0b110,
                      3.0: 0b101,
                      3.3: 0b100,
                      3.6: 0b011,
                      3.9: 0b010,
                      4.2: 0b001,
                      4.5: 0b000}

# Allowed gains
NAU7802_Gain_Values = {1: 0b000,
                       2: 0b001,
                       4: 0b010,
                       8: 0b011,
                       16: 0b100,
                       32: 0b101,
                       64: 0b110,
                       128: 0b111}

# Allowed samples per second
NAU7802_SPS_Values = {10: 0b000,
                      20: 0b001,
                      40: 0b010,
                      80: 0b011,
                      320: 0b111}

# Calibration state
NAU7802_Cal_Status = {'CAL_SUCCESS': 0,
                      'CAL_IN_PROGRESS': 1,
                      'CAL_FAILURE': 2}

NAU7802_VREF = 3.3


class NAU7802(InOut):
  """This class can read values from a NAU7802 load cell conditioner.

  This load cell conditioner is a low-cost 24-bits, single-channel conditioner,
  that can read up to 320 samples per second. It communicates over the I2C
  protocol. The returned value of the InOut is in Volts by default, but can be
  converted to Newtons using the ``gain`` and ``offset`` arguments.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Nau7802 to NAU7802
  """

  def __init__(self,
               i2c_port: int = 1,
               device_address: int = 0x2A,
               gain_hardware: int = 128,
               sample_rate: int = 80,
               int_pin: Optional[int] = None,
               gain: float = 1,
               offset: float = 0) -> None:
    """Checks the validity of the arguments.

    Args:
      i2c_port: The I2C port over which the NAU7802 should communicate. On most
        Raspberry Pi models the default I2C port is `1`.
      device_address: The I2C address of the NAU7802. It is impossible to
        change this address, so it is not possible to have several NAU7802
        connected on the same I2C bus.
      gain_hardware: The gain to be used by the programmable gain amplifier.
        Setting a high gain allows reading small voltages with a better
        precision, but it might saturate the sensor for higher voltages.
        Available gains are:
        ::

          1, 2, 4, 8, 16, 32, 64, 128

      sample_rate: The sample rate for data conversion. The higher the rate,
        the greater the noise. Available sample rates are:
        ::

          10, 20, 40, 80, 320

      int_pin: Optionally, reads the end of conversion signal from the polarity
        of a GPIO rather than from an I2C register. Speeds up the reading and
        decreases the traffic on the bus, but requires one extra wire. Give the
        index of the GPIO in BCM convention, as an :obj:`int`.
      gain: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.
      offset: Allows to tune the output value according to the formula :
        :math:`output = gain * tension + offset`.
      
    .. versionadded:: 1.5.8 *int_pin* argument
    .. versionremoved:: 2.0.0 *backend* and *ft232h_ser_num* arguments
    """

    self._bus = None
    self._int_pin = None

    super().__init__()

    self._bus = smbus2.SMBus(i2c_port)

    self._device_address = device_address

    if gain_hardware not in NAU7802_Gain_Values:
      raise ValueError("gain_hardware should be in {}".format(list(
        NAU7802_Gain_Values.keys())))
    else:
      self._gain_hardware = gain_hardware

    if sample_rate not in NAU7802_SPS_Values:
      raise ValueError("sample_rate should be in {}".format(list(
        NAU7802_SPS_Values.keys())))
    else:
      self._sample_rate = sample_rate

    if int_pin is not None and not isinstance(int_pin, int):
      raise TypeError('int_pin should be an int !')
    self._int_pin = int_pin

    self._gain = gain
    self._offset = offset
    self._retries = 5

  def open(self) -> None:
    """Initializes the I2C communication and the device."""

    if not self._is_connected():
      raise IOError("The NAU7802 is not connected")

    self.log(logging.INFO, "Setting up the NAU7802")

    # Resetting the device
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_RR'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)
    sleep(0.001)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_RR'],
                  NAU7802_Scale_Registers['PU_CTRL'], 0)

    # Powering up the device - takes approx 200µs
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUD'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUA'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)

    t0 = time()
    while not self._get_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUR'],
                            NAU7802_Scale_Registers['PU_CTRL']):
      if time() - t0 > 0.1:
        raise TimeoutError("Waited too long for the device to power up !")

    # Setting the Low Drop Out voltage to 3.3V and setting the gain
    value = NAU7802_Gain_Values[self._gain_hardware]
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['CTRL1'], value)
    self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_AVDDS'],
                  NAU7802_Scale_Registers['PU_CTRL'], 1)

    # Setting the sample rate
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['CTRL2'],
                              NAU7802_SPS_Values[self._sample_rate] << 4)

    # Turning off CLK_CHP
    self._bus.write_byte_data(self._device_address,
                              NAU7802_Scale_Registers['ADC'], 0x30)

    # Enabling 330pF decoupling cap on channel 2
    self._set_bit(NAU7802_PGA_PWR_Bits['PGA_PWR_PGA_CAP_EN'],
                  NAU7802_Scale_Registers['PGA_PWR'], 1)

    # Re-calibrating the analog front-end
    self._set_bit(NAU7802_CTRL2_Bits['CTRL2_CALS'],
                  NAU7802_Scale_Registers['CTRL2'], 1)

    t0 = time()
    while self._cal_afe_status() != NAU7802_Cal_Status['CAL_SUCCESS']:
      if time() - t0 > 1:
        raise TimeoutError("Waited too long for the calibration to occur !")
      if self._cal_afe_status() == NAU7802_Cal_Status['CAL_FAILURE']:
        raise IOError("Calibration failed !")

    if self._int_pin is not None:
      self.log(logging.INFO, "Setting up the GPIOs")
      GPIO.setmode(GPIO.BCM)
      GPIO.setup(self._int_pin, GPIO.IN)

  def get_data(self) -> List[float]:
    """Reads the registers containing the conversion result.

    The output is in Volts by default, and can be converted to Newtons using
    gain and offset.

    Returns:
      A :obj:`list` containing the timeframe and the output value.
    """

    # Waiting for data to be ready
    t0 = time()
    while not self._data_available():
      if time() - t0 > 0.5:
        raise TimeoutError('Waited too long for data to be ready !')

    out = [time()]

    # Reading the output data, and handling I2C errors
    i = 0
    block = None
    while i < self._retries:
      try:
        block = self._bus.read_i2c_block_data(
            self._device_address, NAU7802_Scale_Registers['ADCO_B2'], 3)
        break
      # If an I2C error is caught, retrying to get the value or raising
      except OSError:
        if i == self._retries:
          self.log(logging.ERROR, "Retries count exhausted !")
          raise
        i += 1
        continue

    self.log(logging.DEBUG,
             f"Read {block} from register {NAU7802_Scale_Registers['ADCO_B2']}"
             f" at address {self._device_address}")
    value_raw = (block[0] << 16) | (block[1] << 8) | block[2]

    # Converting raw data into Volts or Newtons
    if block[0] >> 7:
      value_raw -= 2 ** 24
    value = value_raw / 2 ** 23 * 0.5 * NAU7802_VREF / self._gain_hardware
    out.append(self._offset + self._gain * value)
    return out

  def close(self) -> None:
    """Powers down the device."""

    if self._bus is not None:
      # Powering down the device
      self.log(logging.INFO, "Powering down the NAU7802")
      self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUD'],
                    NAU7802_Scale_Registers['PU_CTRL'], 0)
      sleep(0.001)
      self._set_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_PUA'],
                    NAU7802_Scale_Registers['PU_CTRL'], 0)
      self.log(logging.INFO, "Closing the I2C connection to the NAU7802")
      self._bus.close()

    if self._int_pin is not None:
      self.log(logging.INFO, "Cleaning up the GPIOs")
      GPIO.cleanup()

  def _is_connected(self) -> bool:
    """Tries reading a byte from the device.

    Returns:
      :obj:`True` if reading was successful, else :obj:`False`
    """

    try:
      self._bus.read_byte(self._device_address)
      return True
    except IOError:
      return False

  def _data_available(self) -> bool:
    """Returns :obj:`True` if data is available, :obj:`False` otherwise."""

    # EOC signal from the I2C communication
    if self._int_pin is None:
      return self._get_bit(NAU7802_PU_CTRL_Bits['PU_CTRL_CR'],
                           NAU7802_Scale_Registers['PU_CTRL'])
    else:
      return bool(GPIO.input(self._int_pin))

  def _set_bit(self,
               bit_number: int,
               register_address: int,
               bit: int) -> None:
    """Sets a given bit in the specified register.

    Args:
      bit_number: Position of the bit in the register, as an :obj:`int`.
      register_address: Index of the register, as an :obj:`int`.
      bit: Value of the bit, as an :obj:`int`.
    """

    value = self._bus.read_i2c_block_data(self._device_address,
                                          register_address, 1)[0]
    self.log(logging.DEBUG, f"Read {value} from register {register_address}"
                            f" at address {self._device_address}")
    if bit:
      value |= (1 << bit_number)
    else:
      value &= ~(1 << bit_number)
      self.log(logging.DEBUG, f"Writing {value} to register {register_address}"
                              f" at address {self._device_address}")
    self._bus.write_byte_data(self._device_address, register_address, value)

  def _get_bit(self,
               bit_number: int,
               register_address: int) -> bool:
    """Reads a given bit in the specified register.

    Args:
      bit_number: Position of the bit in the register, as an :obj:`int`.
      register_address: Index of the register, as an :obj:`int`.

    Returns:
      :obj:`True` if the bit value is 1, else :obj:`False`.
    """

    value = self._bus.read_i2c_block_data(self._device_address,
                                          register_address, 1)[0]
    self.log(logging.DEBUG, f"Read {value} from register {register_address}"
                            f" at address {self._device_address}")
    value = value >> bit_number & 1
    return bool(value)

  def _cal_afe_status(self) -> int:
    """Reads the calibration status bits.

    Returns:
      The :obj:`int` value corresponding to the current calibration status.
    """

    if self._get_bit(NAU7802_CTRL2_Bits['CTRL2_CAL_ERROR'],
                     NAU7802_Scale_Registers['CTRL2']):
      return NAU7802_Cal_Status['CAL_FAILURE']
    elif self._get_bit(NAU7802_CTRL2_Bits['CTRL2_CALS'],
                       NAU7802_Scale_Registers['CTRL2']):
      return NAU7802_Cal_Status['CAL_IN_PROGRESS']
    else:
      return NAU7802_Cal_Status['CAL_SUCCESS']



# coding: utf-8

"""The name "Biotens" has been used for years before being replaced by
"JVLMac140". Keeping an alias here to make user's life easier."""

from ..actuator import JVLMac140


class Biotens(JVLMac140):
  """Simply an alias of :class:`~crappy.actuator.JVLMac140` for convenience for
   the users at the LaMcube laboratory."""



# coding: utf-8

from typing import Dict, Union, Any, Optional, Iterable, List, Tuple
from time import time, sleep
from subprocess import Popen, PIPE, STDOUT, TimeoutExpired
from threading import Thread
from queue import Queue, Empty
from ast import literal_eval
from pickle import loads, dumps, UnpicklingError
from socket import timeout, gaierror
from itertools import chain
import logging

from .meta_block import Block
from .._global import OptionalModule

try:
  import paho.mqtt.client as mqtt
except (ModuleNotFoundError, ImportError):
  mqtt = OptionalModule("paho-mqtt")

TopicsType = Iterable[Union[str, Iterable[str]]]


class ClientServer(Block):
  """This Block can exchange data on a local network using the MQTT protocol.
  
  It communicates with an MQTT broker, from which it can receive data by
  subscribing to topics, and to which it can send data by publishing in topics.
  This Block can also manage the execution of a Mosquitto broker, so that the
  broker doesn't need to be manually started before running the Crappy script.

  This Block is intended for communication with remote devices, that may run
  Crappy scripts or other scripts handling data with the correct syntax.
  Potential uses include acquisition from battery-powered devices (e.g.
  microcontrollers) acquiring data in enclosed areas or rotating parts, data
  transfer between machines to delegate processing, communication with other
  programs on a same machine, etc.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Client_server to ClientServer
  """

  def __init__(self,
               broker: bool = False,
               address: str = 'localhost',
               port: int = 1883,
               init_output: Optional[Dict[str, Any]] = None,
               topics: Optional[TopicsType] = None,
               cmd_labels: Optional[TopicsType] = None,
               labels_to_send: Optional[TopicsType] = None,
               display_freq: bool = False,
               freq: Optional[float] = 200,
               spam: bool = False,
               debug: Optional[bool] = False) -> None:
    """Checks the validity of the arguments and sets the instance attributes.

    Args:
      broker: If :obj:`True`, starts the Mosquitto broker during the prepare
        loop and stops it during the finish loop. If Mosquitto is not installed
        a :exc:`FileNotFoundError` is raised.
      address: The network address on which the MQTT broker is running, as a
        :obj:`str`.
      port: A network port on which the MQTT broker is listening, as an 
        :obj:`int`.
      init_output: A :obj:`dict` containing for each label in ``topics`` the 
        first value to be sent in the output Links. Should be given in case the 
        data comes from several sources and data for all labels may not be 
        available during the first loops. Must also be given is ``spam`` is set
        to :obj:`True`.
      topics: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        :obj:`str` and/or iterables of :obj:`str`. Each string corresponds to
        the name of a label in Crappy. Each element in the iterable (string or
        iterable of strings) is considered to be the name of an MQTT topic, to
        which the client subscribes. After a message has been received on that
        topic, the Block returns for each label in the topic (just the given
        string or each string in the iterable) the corresponding data from the
        message. It also returns the current timestamp in the label `'t(s)'`.
      cmd_labels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        :obj:`str` and/or iterables of :obj:`str`. Each string corresponds to
        the name of a label in Crappy. Each element in the iterable (string or
        iterable of strings) is considered to be the name of an MQTT topic, in
        which the client publishes. Grouping labels in a same topic (i.e.
        strings in a same iterable) allows to keep the synchronization between
        signals coming from a same Block, as they will be published together in
        a same message. This is mostly useful for sending a signal along with
        its timeframe.
      labels_to_send: An iterable (like a :obj:`list` or a :obj:`tuple`)
        containing :obj:`str` and/or iterables of :obj:`str`. Allows to rename
        the labels before publishing data. The structure of ``labels_to_send``
        should be the exact same as ``cmd_labels``, with each label in
        ``labels_to_send`` replacing the corresponding one in ``cmd_labels``.
        This is especially useful for transferring several signals along with
        their timestamps, as the label ``'t(s)'`` should not appear more than
        once in the topics.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      freq: The target looping frequency for the Block. If :obj:`None`, loops
        as fast as possible.
      spam: If :obj:`True`, sends the last received values at each loop even if
        no new values were received from the broker. When set to :obj:`True`,
        the ``init_output`` must be provided.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    Note:
      - ``broker``:
        In order for the Block to run, an MQTT broker must be running at the
        specified address on the specified port. If not, an
        :exc:`ConnectionRefusedError` is raised. The broker can be started and
        stopped manually by the user independently of the execution of Crappy.
        It also doesn't need to be Mosquitto, any other MQTT broker can be
        used.

      - ``topics``:
        The presence of the same label in multiple topics will most likely lead
        to a data loss.

      - ``cmd_labels``:
        It is not possible to group signals coming from different Blocks in a
        same topic.

      - ``labels_to_send``:
        Differences in the structure of ``labels_to_send`` and ``cmd_labels``
        will not always raise an error, but may lead to a data loss.

      - **Single-value iterables**:
        Single-value iterables can be shortened as strings.
        ::

          topics=[('cmd1',), ('cmd2',)]
          cmd_labels=[('cmd1',), ('cmd2',)]
          labels_to_send=[('cmd1',), ('cmd2',)]

        is equivalent to
        ::

          topics=['cmd1', 'cmd2']
          cmd_labels=['cmd1', 'cmd2']
          labels_to_send=['cmd1', 'cmd2']


    Examples:
      - ``topics``:
        If
        ::

          topics=[('t1', 'cmd1'), 'sign']

        the client will subscribe to the topics
        ::

          ('t1', 'cmd1')
          ('sign',)

        The Block will return data associated with the labels
        ::

          't1', 'cmd1'
          'sign'


      - ``cmd_labels``:
        If
        ::

          cmd_labels=[('t1', 'cmd1'), 'sign']

        the client will publish data in the form of
        ::

          [[t1_0, cmd1_0], [t1_1, cmd1_1], ...]
          [[sign_0], [sign_1], ...]

        in the topics
        ::

           ('t1', 'cmd1')
           ('sign',)


      - ``labels_to_send``:
        If
        ::

          cmd_labels=[('t(s)', 'cmd'), 'sign']
          labels_to_send=[('t1', 'cmd1'), 'sign']

        the data from labels
        ::

          't(s)', 'cmd'

        will be published in the topic
        ::

          ('t1', 'cmd1')

        and the data from label
        ::

          'sign'

        in the topic
        ::

          ('sign',)

    .. versionadded:: 1.5.10 *verbose*, *freq* and *spam* arguments
    .. versionadded:: 2.0.0 *debug* argument
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    self._client: Optional[mqtt.Client] = None
    self._reader: Optional[Thread] = None
    self._proc: Optional[Popen] = None

    super().__init__()
    self.niceness = -10
    self.display_freq = display_freq
    self.freq = freq
    self.debug = debug

    # Setting the args
    self._broker = broker
    self._address = address
    self._port = port
    self._spam = spam
    self._init_output = init_output if init_output is not None else dict()

    self._stop_mosquitto = False
    
    # These attributes may be set later
    self._topics: Optional[List[Tuple[str, ...]]] = None
    self._last_out_val: Dict[str, Any] = dict()
    self._buffer_output: Optional[Dict[Tuple[str, ...], Queue]] = None
    self._cmd_labels: Optional[List[Tuple[str, ...]]] = None
    self._labels_to_send: Optional[List[Tuple[str, ...]]] = None

    if topics is None and cmd_labels is None:
      self.log(logging.WARNING, "The Client-server Block is neither an input "
                                "nor an output !")

    # Preparing for receiving data
    if topics is not None:
      # Replacing strings with tuples
      self._topics = [(topic,) if isinstance(topic, str) else tuple(topic) for
                      topic in topics]

      # The last out vals are given for each label, not each topic
      self._last_out_val = {label: None for label in chain(*self._topics)}

    # Preparing for publishing data
    if cmd_labels is not None:
      # Replacing strings with tuples
      self._cmd_labels = [(topic,) if isinstance(topic, str) else tuple(topic)
                          for topic in cmd_labels]

      if labels_to_send is not None:
        # Replacing strings with tuples
        labels_to_send = [(topic,) if isinstance(topic, str) else tuple(topic)
                          for topic in labels_to_send]

        # Making sure the labels to send have the correct syntax
        if len(labels_to_send) != len(self._cmd_labels):
          raise ValueError("Either a label_to_send should be given for "
                           "every cmd_label, or none should be given ")

        # Preparing to rename labels to send using a dictionary
        self._labels_to_send = {cmd_label: label_to_send for
                                cmd_label, label_to_send in
                                zip(self._cmd_labels, labels_to_send)}

  def prepare(self) -> None:
    """Starts the broker and connects to it."""

    # Making sure the necessary inputs and outputs are present
    if self._topics is not None and not self.outputs:
      raise ValueError("topics are specified but there's no output link !")
    if self._cmd_labels is not None and not self.inputs:
      raise ValueError("cmd_labels are specified but there's no input link !")

    # Setting the buffer here because Queue objects cannot be set during
    # __init__ in spawn multiprocessing mode
    if self._topics is not None:
      # The buffer for received data is a dictionary of queues
      self._buffer_output = {topic: Queue() for topic in self._topics}

    # Starting the broker
    if self._broker:
      self.log(logging.INFO, f"Starting the Mosquitto broker on port "
                             f"{self._port}")
      self._launch_mosquitto()
      # Creating and starting a Thread reading the stdout of the broker
      self._reader = Thread(target=self._output_reader)
      self._reader.start()
      sleep(2)
      self.log(logging.INFO, "Waiting for Mosquitto to start")
      sleep(2)

    # Instantiating the client here as it cannot be set during __init__ in
    # spawn multiprocessing mode
    self._client = mqtt.Client(str(time()))
    self._client.on_connect = self._on_connect
    self._client.on_message = self._on_message
    self._client.reconnect_delay_set(max_delay=10)

    # Connecting to the broker
    try_count = 15
    while True:
      try:
        self.log(logging.INFO, f"Connecting to the broker at address "
                               f"{self._address} on port {self._port}")
        self._client.connect(self._address, port=self._port, keepalive=10)
        break
      except timeout:
        raise TimeoutError("Impossible to reach the given address, aborting")
      except gaierror:
        raise ValueError("Invalid address given, please check the spelling")
      except ConnectionRefusedError:
        try_count -= 1
        if try_count == 0:
          raise ConnectionRefusedError("Connection refused, the broker may not"
                                       " be running or you may not have the "
                                       "rights to connect")
        sleep(1)

    self.log(logging.INFO, "Starting the client loop")
    self._client.loop_start()

  def loop(self) -> None:
    """Receives data from the broker and/or sends data to the broker.

    The received data is then sent to the crappy Blocks connected to this one.
    """

    """Loop for receiving data
    Each queue in the buffer is checked once: if not empty then the first list 
    of data is popped. This data is then associated to the corresponding 
    labels in dict_out. dict_out is finally returned if not empty. All the 
    labels should be returned at each loop iteration, so a buffer stores the 
    last value for each label and returns it if no other value was received. In 
    case no value was received yet for a given label, the user can also provide 
    init values to be sent at the beginning."""
    if self._topics is not None:
      dict_out = dict()
      for topic in self._buffer_output:
        if not self._buffer_output[topic].empty():
          try:
            data_list = self._buffer_output[topic].get_nowait()
            for label, data in zip(topic, data_list):
              dict_out[label] = data
          except Empty:
            pass
      # Updating the _last_out_val buffer, and completing dict_out before
      # sending data if necessary
      if dict_out or self._spam:
        for topic in self._buffer_output:
          for label in topic:
            if label not in dict_out:
              if self._last_out_val[label] is not None:
                dict_out[label] = self._last_out_val[label]
              elif label in self._init_output:
                dict_out[label] = self._init_output[label]
              else:
                raise ValueError(f"No value received for the topic {label} and"
                                 f" no init value given !")
            else:
              self._last_out_val[label] = dict_out[label]
        # Adding the timestamp before sending
        dict_out['t(s)'] = time() - self.t0
        self.send(dict_out)

    """Loop for sending data
    Data is first received as a list of dictionaries. For each topic, trying to 
    find a dictionary containing all the corresponding labels. Once this 
    dictionary has been found, its data is published as a list of list of 
    values."""
    if self._cmd_labels is not None:
      received_data = self.recv_all_data_raw()
      for topic in self._cmd_labels:
        for dic in received_data:
          if dic is not None and all(label in dic for label in topic):
            if self._labels_to_send is not None:
              topic_to_send = self._labels_to_send[topic]
            else:
              topic_to_send = topic
            self._client.publish(
              topic=str(topic_to_send),
              payload=dumps([dic[label] for label in topic]),
              qos=0)
            self.log(logging.DEBUG, f"Sent {[dic[label] for label in topic]}"
                                    f"on the topic {topic_to_send} with QOS 0")
            break

  def finish(self) -> None:
    """Disconnects from the broker and stops it."""

    # Disconnecting from the broker
    if self._client is not None:
      self.log(logging.INFO, "Stopping the client loop")
      self._client.loop_stop()
      self.log(logging.INFO, "Disconnecting from the broker")
      self._client.disconnect()

    # Stopping the broker
    if self._broker and self._proc is not None:
      try:
        self.log(logging.INFO, "Stopping the Mosquitto broker")
        self._proc.terminate()
        self._proc.wait(timeout=15)
        self.log(logging.INFO, f"Mosquitto terminated with return code "
                               f"{self._proc.returncode}")
        self._stop_mosquitto = True
        if self._reader is not None:
          self._reader.join(0.2)
          if self._reader.is_alive():
            self.log(logging.WARNING, "Reader thread failed to stop !")

      except TimeoutExpired:
        self.log(logging.WARNING, "Mosquitto did not terminate in time, "
                                  "killing it")
        self._proc.kill()

  def _launch_mosquitto(self) -> None:
    """Starts Mosquitto in a subprocess."""

    try:
      self._proc = Popen(['mosquitto', '-p', str(self._port)],
                         stdout=PIPE,
                         stderr=STDOUT)
    except FileNotFoundError:
      raise FileNotFoundError("Mosquitto is not installed !")

  def _output_reader(self) -> None:
    """Reads the output strings from Mosquitto's subprocess."""

    while not self._stop_mosquitto:
      for line in iter(self._proc.stdout.readline, b''):
        self.log(logging.INFO, f"[Mosquitto] {line.decode()}")
        if 'Error: Address already in use' in line.decode():
          self.log(logging.WARNING, "Mosquitto is already running on this "
                                    "port !")
      sleep(0.1)

  def _on_message(self, _, __, message) -> None:
    """Buffers the received data.

    The received message consists in a list of lists of values. Data is placed
    in the right buffer according to the topic, in the form of lists of values.
    """

    try:
      self.log(logging.DEBUG, f"Received message from the broker: {message}")
      for data_points in zip(*loads(message.payload)):
        self._buffer_output[literal_eval(message.topic)].put_nowait(
          list(data_points))
    except UnpicklingError:
      self.log(logging.WARNING, f"Received message caused UnpicklingError, "
                                f"ignoring it: {message}")

  def _on_connect(self, _, __, ___, rc: Any) -> None:
    """Automatically subscribes to the topics when connecting to the broker."""

    self.log(logging.INFO, f"Connected to the broker with return code {rc}")

    # Subscribing on connect, so that it automatically resubscribes when
    # reconnecting after a connection loss
    if self._topics is not None:
      for topic in self._topics:
        self._client.subscribe(topic=str(topic), qos=0)
        self.log(logging.INFO, f"Subscribed to topic {topic}")

    self._client.loop_start()



# coding: utf-8

from typing import Optional
import numpy as np
import logging
import logging.handlers
from time import sleep

from .camera_process import CameraProcess
from ...tool.image_processing import DICVETool
from ...tool.camera_config import SpotsBoxes


class DICVEProcess(CameraProcess):
  """This :class:`~crappy.blocks.camera_processes.CameraProcess` can perform
  video-extensometry by tracking patches on images using various Digital Image 
  Correlation techniques.
  
  It is used by the :class:`~crappy.blocks.DICVE` Block to parallelize the
  image processing and the image acquisition. It delegates most of the 
  computation to the :class:`~crappy.tool.image_processing.DICVETool`. It is
  from this class that the output values are sent to the downstream Blocks, and
  that the :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` are sent
  to the :class:`~crappy.blocks.camera_processes.Displayer` CameraProcess for
  display.
  
  .. versionadded:: 2.0.0
  """

  def __init__(self,
               patches: SpotsBoxes,
               method: str = 'Disflow',
               alpha: float = 3,
               delta: float = 1,
               gamma: float = 0,
               finest_scale: int = 1,
               iterations: int = 1,
               gradient_iterations: int = 10,
               patch_size: int = 8,
               patch_stride: int = 3,
               border: float = 0.2,
               safe: bool = True,
               follow: bool = True,
               raise_on_exit: bool = True) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      patches: An instance of the
        :class:`~crappy.tool.camera_config.config_tools.SpotsBoxes` class,
        containing the coordinates of the patches to track. This argument is
        passed to the :obj:`~crappy.tool.image_processing.DICVETool` and not
        used in this class.
      method: The method to use to calculate the displacement. `Disflow` uses
        opencv's DISOpticalFlow and `Lucas Kanade` uses opencv's
        calcOpticalFlowPyrLK, while all other methods are based on a basic
        cross-correlation in the Fourier domain. `Pixel precision` calculates
        the displacement by getting the position of the maximum of the
        cross-correlation, and has thus a 1-pixel resolution. It is mainly
        meant for debugging. `Parabola` refines the result of
        `Pixel precision` by interpolating the neighborhood of the maximum, and
        has thus a sub-pixel resolution. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      alpha: Weight of the smoothness term in DISFlow, as a :obj:`float`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      delta: Weight of the color constancy term in DISFlow, as a :obj:`float`.
        This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      gamma: Weight of the gradient constancy term in DISFlow , as a
        :obj:`float`. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      finest_scale: Finest level of the Gaussian pyramid on which the flow
        is computed in DISFlow (`0` means full scale), as an :obj:`int`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      iterations: Maximum number of gradient descent iterations in the
        patch inverse search stage in DISFlow, as an :obj:`int`. This argument
        is passed to the :obj:`~crappy.tool.image_processing.DICVETool` and not
        used in this class.
      gradient_iterations: Maximum number of gradient descent iterations
        in the patch inverse search stage in DISFlow, as an :obj:`int`. This
        argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      patch_size: Size of an image patch for matching in DISFlow
        (in pixels). This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      patch_stride: Stride between neighbor patches in DISFlow. Must be
        less than patch size. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      border: Crop the patch on each side according to this value before
        calculating the displacements. 0 means no cropping, 1 means the entire
        patch is cropped. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      safe: If :obj:`True`, checks whether the patches aren't exiting the
        image, and raises an error if that's the case. This argument is passed
        to the :obj:`~crappy.tool.image_processing.DICVETool` and not used in
        this class.
      follow: It :obj:`True`, the patches will move to follow the displacement
        of the image. This argument is passed to the
        :obj:`~crappy.tool.image_processing.DICVETool` and not used in this
        class.
      raise_on_exit: If :obj:`True`, raises an exception and stops the test
        when losing the patches. Otherwise, simply stops processing but lets 
        the test go on.
    """

    super().__init__()

    # Arguments to pass to the DICVETool
    self._patches = patches
    self._method = method
    self._alpha = alpha
    self._delta = delta
    self._gamma = gamma
    self._finest_scale = finest_scale
    self._iterations = iterations
    self._gradient_iterations = gradient_iterations
    self._patch_size = patch_size
    self._patch_stride = patch_stride
    self._border = border
    self._safe = safe
    self._follow = follow
    
    # Other attributes
    self._raise_on_exit = raise_on_exit
    self._disve: Optional[DICVETool] = None
    self._img0_set = False
    self._lost_patch = False

  def init(self) -> None:
    """Instantiates the :obj:`~crappy.tool.image_processing.DICVETool` that
    will perform the image correlation."""

    self.log(logging.INFO, "Instantiating the Disve tool")
    self._disve = DICVETool(patches=self._patches,
                            method=self._method,
                            alpha=self._alpha,
                            delta=self._delta,
                            gamma=self._gamma,
                            finest_scale=self._finest_scale,
                            iterations=self._iterations,
                            gradient_iterations=self._gradient_iterations,
                            patch_size=self._patch_size,
                            patch_stride=self._patch_stride,
                            border=self._border,
                            safe=self._safe,
                            follow=self._follow)

  def loop(self) -> None:
    """This method grabs the latest frame and gives it for processing to the
    :obj:`~crappy.tool.image_processing.DICVETool`. Then sends the result of
    the correlation to the downstream Blocks.
    
    If there's no new frame grabbed, or if the patches were already lost, 
    doesn't do anything. On the first acquired frame, does not process it but 
    initializes the DICVETool with it instead. Also sends the current patches 
    for display to the :class:`~crappy.blocks.camera_processes.Displayer`
    CameraProcess.
    """

    # Do nothing if the patches were already lost
    if not self._lost_patch:
      try:

        # On the first frame, initialize the correlation
        if not self._img0_set:
          self.log(logging.INFO, "Setting the reference image")
          self._disve.set_img0(np.copy(self.img))
          self._img0_set = True
          return

        # Calculating the displacement and sending it to downstream Blocks
        self.log(logging.DEBUG, "Processing the received image")
        data = self._disve.calculate_displacement(self.img)
        self.send([self.metadata['t(s)'], self.metadata, *data])

        # Sending the patches to the Displayer for display
        self.send_to_draw(self._disve.patches)

      # If the patches are lost, deciding whether to raise exception or not
      except RuntimeError as exc:
        if self._raise_on_exit:
          self._logger.exception("Patch exiting the ROI !", exc_info=exc)
          raise
        self._lost_patch = True
        self.log(logging.WARNING, "Patch exiting the ROI, not processing "
                                  "data anymore !")
    
    # If the patches are lost, sleep to avoid spamming the CPU in vain
    else:
      self.fps_count -= 1
      sleep(0.1)



# coding: utf-8

import logging
import numpy as np
from typing import Dict, Optional, Iterable, Union
from collections import defaultdict

from .meta_block import Block


class Multiplexer(Block):
  """This Block takes data from upstream Blocks as input and interpolates it to
  output all the labels in a common time basis.

  It can take any number of inputs, provided that they all share a common time
  label. It is also possible to choose which labels are considered for
  multiplexing and which are dropped. The interpolation is performed using the
  :obj:`numpy.interp` method.

  This Block is useful for synchronizing data acquired from different sensors,
  e.g. to plot a real-time stress-strain curve with position data coming from a
  :class:`~crappy.blocks.Machine` Block and force data coming from a
  :class:`~crappy.blocks.IOBlock` Block. Multiplexing is however quite
  resource-consuming, so it is preferable to perform interpolation when
  post-processing if real-time is not needed.

  Note:
    This Block doesn't truly output data in real-time as it needs to wait for
    data from all the upstream Blocks before performing the interpolation. It
    should only be used with care as an input to a decision-making Block. This
    is especially true when the upstream Blocks have very different sampling
    rates.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Multiplex to Multiplexer
  """

  def __init__(self,
               time_label: str = 't(s)',
               out_labels: Optional[Union[str, Iterable[str]]] = None,
               interp_freq: float = 200,
               freq: Optional[float] = 50,
               display_freq: bool = False,
               debug: Optional[bool] = False) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      time_label: The label carrying the time information.
      out_labels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        the labels to multiplex, except for the time label that is given
        separately in the ``time_label`` argument. The Block also doesn't
        output anything until data has been received on all these labels. If
        left to :obj:`None`, all the received labels are considered. **It is
        recommended to always set this argument !** It is also possible to
        give this argument as a single :obj:`str` (i.e. not in an iterable),
        although multiplexing a single label is of limited interest.
      freq: The target looping frequency for the Block. If :obj:`None`, loops 
        as fast as possible.
      display_freq: If :obj:`True`, displays the looping frequency of the
        Block.
      debug: If :obj:`True`, displays all the log messages including the
        :obj:`~logging.DEBUG` ones. If :obj:`False`, only displays the log
        messages with :obj:`~logging.INFO` level or higher. If :obj:`None`,
        disables logging for this Block.

    .. versionadded:: 1.5.9 *verbose* argument
    .. versionchanged:: 1.5.10 renamed *key* argument to *time_label*
    .. versionadded 2.0.0 *debug*, *out_labels* and *interp_freq* arguments
    .. versionchanged:: 2.0.0 renamed *verbose* argument to *display_freq*
    """

    super().__init__()
    self.freq = freq
    self.display_freq = display_freq
    self.debug = debug

    # Initializing the attributes
    self._time_label = time_label
    self._interp_freq = interp_freq
    self._data: Dict[str, np.ndarray] = defaultdict(self._default_array)
    self._delta: float = 1 / self._interp_freq / 20
    self._last_max_t: float = -float('inf')

    # Forcing the out_labels into a list
    if out_labels is not None and isinstance(out_labels, str):
      self._out_labels = [out_labels]
    elif out_labels is not None:
      self._out_labels = list(out_labels)
    else:
      self._out_labels = None

  def loop(self) -> None:
    """Receives data, interpolates it, and sends it to the downstream
    Blocks."""

    # Receiving all the upcoming data
    data = self.recv_all_data_raw()

    # Iterating over all the links
    for link_data in data:
      # Only data associated with a time label can be multiplexed
      if self._time_label not in link_data:
        continue
      # Extracting the time information from the data
      timestamps = link_data.pop(self._time_label)

      # Adding data from each label in the buffer
      for label, values in link_data.items():
        # Only the labels specified in out_labels is considered
        if self._out_labels is not None and label not in self._out_labels:
          continue

        # Adding the received values to the buffered ones
        self._data[label] = np.concatenate((self._data[label],
                                            np.array((timestamps, values))),
                                           axis=1)
        # Sorting the buffered data, if a same label comes from multiple Links
        self._data[label] = self._data[label][
          :, self._data[label][0].argsort()]

    # Aborting if there's no data to process
    if not self._data:
      self.log(logging.DEBUG, "No data in the buffer to process")
      return

    # Making sure there's data for all the requested labels
    if (self._out_labels is not None and
        any(label not in self._data for label in self._out_labels)):
      self.log(logging.DEBUG, "Not all the requested labels received yet")
      return

    # There should also be at least two values for each label
    if any(len(self._data[label][0]) < 2 for label in self._data):
      self.log(logging.DEBUG, "Not at least 2 values for each label in buffer")
      return

    # The two values should also be separated by at least one time period
    if any(np.ptp(self._data[label][0]) < 1 / self._interp_freq
           for label in self._data):
      self.log(logging.DEBUG, "At least one label has values too close "
                              "together compared to interpolation frequency")
      return

    # Getting the minimum time for the interpolation (maximin over all labels)
    min_t = max(data[0, 0] for data in self._data.values())
    # The minimum must be higher than the previous maximum
    min_t = max(min_t, self._last_max_t + self._delta)
    # Correcting to the closest upper multiple of the time interval
    min_t = min_t + (1 / self._interp_freq) - min_t % (1 / self._interp_freq)

    # Getting the maximum time for the interpolation (minimax over all labels)
    max_t = min(data[0, -1] for data in self._data.values())
    # Correcting to the closest lower multiple of the time interval
    max_t = max_t - (1 / self._interp_freq) + max_t % (1 / self._interp_freq)

    if max_t < min_t:
      self.log(logging.DEBUG, "Ranges not matching for interpolation")
      return

    # The array containing the timestamps for interpolating
    interp_times = np.arange(min_t, max_t + self._delta, 1 / self._interp_freq)

    # Making sure there are points to interpolate
    if not np.any(interp_times):
      self.log(logging.DEBUG, "No time points for interpolation")
      return

    to_send = dict()
    self._last_max_t = max_t

    # Building the dict of values to send
    for label, values in self._data.items():
      to_send[label] = list(np.interp(interp_times, values[0], values[1]))
      # Keeping the last data point before max_t to pass this information on
      last = values[:, values[0] <= max_t][:, -1]
      # Removing the used values from the buffer, except the last data point
      self._data[label] = np.column_stack((last, values[:, values[0] > max_t]))

    if to_send:
      # Adding the time values to the dict of values to send
      to_send[self._time_label] = list(interp_times)

      # Sending the values
      for i, _ in enumerate(interp_times):
        self.send({label: values[i] for label, values in to_send.items()})

  @staticmethod
  def _default_array() -> np.ndarray:
    """Helper function for the default dict."""

    return np.array(([], []))



# coding: utf-8

import numpy as np
from typing import Dict, Any, Optional
import logging

from .meta_modifier import Modifier


class Median(Modifier):
  """Modifier waiting for a given number of data points to be received, then
  returning their median, and starting all over again.

  Unlike :class:`~crappy.modifier.MovingMed`, it only returns a value once
  every ``n_points`` points.
  
  .. versionadded:: 1.4.0
  """

  def __init__(self, n_points: int = 100) -> None:
    """Sets the args and initializes the parent class.

    Args:
      n_points: The number of points on which to compute the median.

    .. versionchanged:: 1.5.10 renamed *npoints* argument to *n_points*
    """

    super().__init__()
    self._n_points = n_points
    self._buf = None

  def __call__(self, data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Receives data from the upstream Block, and computes the median of every
    label once the right number of points have been received. Then empties the
    buffer and returns the medians.

    If there are not enough points, doesn't return anything.
    
    .. versionchanged:: 2.0.0 renamed from evaluate to __call__
    """

    self.log(logging.DEBUG, f"Received {data}")

    # Initializing the buffer
    if self._buf is None:
      self._buf = {key: [value] for key, value in data.items()}

    ret = {}
    for label in data:
      # Updating the buffer with the newest data
      self._buf[label].append(data[label])

      # Once there's enough data in the buffer, calculating the median value
      if len(self._buf[label]) == self._n_points:
        try:
          ret[label] = np.median(self._buf[label])
        except TypeError:
          ret[label] = self._buf[label][-1]

        # Resetting the buffer
        self._buf[label].clear()

    if ret:
      self.log(logging.DEBUG, f"Sending {ret}")
      return ret

    self.log(logging.DEBUG, "Not returning any data")



# coding: utf-8

from typing import Dict, Type

from .adafruit_dc_motor_hat import DCMotorHat
from .fake_dc_motor import FakeDCMotor
from .fake_stepper_motor import FakeStepperMotor
from .jvl_mac_140 import JVLMac140
from .kollmorgen_servostar_300 import ServoStar300
from .newport_tra6ppd import NewportTRA6PPD
from .oriental_ard_k import OrientalARDK
from .pololu_tic import PololuTic
from .schneider_mdrive_23 import SchneiderMDrive23

from .ft232h import DCMotorHatFT232H

from .meta_actuator import MetaActuator, Actuator

from ._deprecated import deprecated_actuators
actuator_dict: Dict[str, Type[Actuator]] = MetaActuator.classes



# coding: utf-8

from ..._global import DefinitionError


class MetaCamera(type):
  """Metaclass ensuring that two cameras don't have the same name, and that all
  cameras define the required methods. Also keeps track of all the Camera
  classes, including the custom user-defined ones.

  .. versionadded:: 1.4.0
  .. versionchanged:: 1.5.10
     not checking anymore for mandatory methods in :meth:`__init__`
  """

  classes = {}

  def __new__(mcs, name: str, bases: tuple, dct: dict) -> type:
    return super().__new__(mcs, name, bases, dct)

  def __init__(cls, name: str, bases: tuple, dct: dict) -> None:
    super().__init__(name, bases, dct)

    # Checking that a Camera with the same name doesn't already exist
    if name in cls.classes:
      raise DefinitionError(f"The {name} class is already defined !")

    # Otherwise, saving the class
    cls.classes[name] = cls



# coding: utf-8

deprecated_inouts = {'Ads1115': 'ADS1115',
                     'Daqmx': 'DAQmx',
                     'Fake_inout': 'FakeInOut',
                     'Gpio_pwm': 'GPIOPWM',
                     'Gpio_switch': 'GPIOSwitch',
                     'Gsm': 'Sim868',
                     'Koll': 'KollmorgenAKDPDMM',
                     'Labjack_t7': 'LabjackT7',
                     'Labjack_ue9': 'LabjackUE9',
                     'Mcp9600': 'MCP9600',
                     'Mprls': 'MPRLS',
                     'Nau7802': 'NAU7802',
                     'Nidaqmx': 'NIDAQmx',
                     'Opsens': 'HandySens',
                     'Pijuice': 'PiJuice',
                     'Spectrum': 'SpectrumM2I4711',
                     'T7_streamer': 'T7Streamer',
                     'Waveshare_ad_da': 'WaveshareADDA',
                     'Waveshare_high_precision': 'WaveshareHighPrecision'}



# coding: utf-8

deprecated_actuators = {'CM_drive': 'SchneiderMDrive23',
                        'Fake_motor': 'FakeDCMotor',
                        'Motorkit_pump': 'DCMotorHat',
                        'Oriental': 'OrientalARDK',
                        'Pololu_tic': 'PololuTic',
                        'Servostar': 'ServoStar300',
                        'Tra6ppd': 'NewportTRA6PPD'}



# coding: utf-8

from time import time
from typing import Optional, List, Iterable
from dataclasses import dataclass
import logging
from  warnings import warn

from .meta_inout import InOut
from .._global import OptionalModule

try:
  from ue9 import UE9
except (ModuleNotFoundError, ImportError):
  UE9 = OptionalModule("LabJackPython")


@dataclass
class _Channel:
  """This class is a simple structure holding all the attributes a Labjack UE9
  channel can have."""

  num: int
  range_num: int = 12
  gain: float = 1
  offset: float = 0
  make_zero: bool = False


class LabjackUE9(InOut):
  """This class can read the analog input channels from a Labjack UE9 device.

  It cannot read nor drive any of the other inout or output channels on the
  UE9. The UE9 model has been discontinued, and replaced by the T7 model (see
  :class:`~crappy.inout.LabjackT7`). It is thus likely that this class won't be
  further improved in the future.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed from Labjack_ue9 to LabjackUE9
  """

  def __init__(self,
               channels: Iterable[int],
               gain: Optional[Iterable[float]] = None,
               offset: Optional[Iterable[float]] = None,
               make_zero: Optional[Iterable[bool]] = None,
               resolution: Optional[Iterable[int]] = None) -> None:
    """Sets the arguments and initializes the parent class.

    Args:
      channels: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        all the channels to read, given as :obj:`int`. Only the `AIN` channels
        can be read by this class, so to read the channel `AIN2` the integer
        `2` should be added to the iterable.
      gain: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each channel the gain to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows :
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no gain is applied to the measured values.
      offset: An iterable (like a :obj:`list` or a :obj:`tuple`) containing for
        each channel the offset to apply to the measured voltage, as a
        :obj:`float`. The returned voltage is calculated as follows :
        :math:`returned\_voltage = gain * measured\_voltage + offset`. If not
        given, no offset is applied to the measured values.
      make_zero: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each channel a :obj:`bool` indicating whether the channel should be
        zeroed or not. If so, data will be acquired on this channel before the
        test starts, and a compensation value will be deduced so that the
        offset of this channel is `0`. **It will only take effect if the**
        ``make_zero_delay`` **argument of the** :class:`~crappy.blocks.IOBlock`
        **controlling the Labjack is set** ! If not given, the channels are by
        default not zeroed.
      resolution: An iterable (like a :obj:`list` or a :obj:`tuple`) containing
        for each channel the resolution of the acquisition as an integer. Refer
        to Labjack documentation for more details. The higher this value the
        better the resolution, but the lower the speed. The possible range is
        `1` to `12`, and the default is `12`.

    Note:
      All the :iterables given as arguments for the channels should have the
      same length. If that's not the case, all the given iterables are treated
      as if they had the same length as the shortest given one.
    """

    warn(f"Starting from version 2.1.0, {type(self).__name__} will be moved "
         f"to crappy.collection. Your code that uses it will still work as "
         f"is, except you will now need to import crappy.collection at the "
         f"top of your script.", FutureWarning)

    self._handle = None

    super().__init__()

    # Setting the defaults for arguments that are not given
    if resolution is None:
      resolution = [12 for _ in channels]
    if gain is None:
      gain = [1 for _ in channels]
    if offset is None:
      offset = [0 for _ in channels]
    if make_zero is None:
      make_zero = [False for _ in channels]

    self._channels = [_Channel(num=chan, range_num=r_num, gain=g,
                               offset=off, make_zero=make_z)
                      for chan, r_num, g, off, make_z in
                      zip(channels, resolution, gain, offset, make_zero)]
    self.log(logging.DEBUG, f"Input channels: {self._channels}")

  def open(self) -> None:
    """Opens the connection to the Labjack."""

    self.log(logging.INFO, "Opening the connection to the Labjack")
    self._handle = UE9()

  def make_zero(self, delay: float) -> None:
    """Overriding the method of the parent class, because the user can choose
    which channels should be zeroed or not.

    It simply performs the regular zeroing, and resets the compensation to
    zero for the channels that shouldn't be zeroed.
    
    .. versionadded:: 1.5.10
    """

    # No need to acquire data if no channel should be zeroed
    if any(chan.make_zero for chan in self._channels):

      # Acquiring the data
      super().make_zero(delay)

      # Proceed only if the acquisition went fine
      if self._compensations:
        # Resetting the compensation for channels that shouldn't be zeroed
        self._compensations = [comp if chan.make_zero else 0 for comp, chan
                               in zip(self._compensations, self._channels)]

  def get_data(self) -> List[float]:
    """Reads sequentially the channels and returns the acquired values,
    corrected by the given gains and offsets."""

    return [time()] + [self._handle.getAIN(chan.num, Resolution=chan.range_num)
                       * chan.gain + chan.offset for chan in self._channels]

  def close(self) -> None:
    """Closes the connection to the Labjack, if it was already opened."""

    if self._handle is not None:
      self.log(logging.INFO, "Closing the connection to the Labjack")
      self._handle.close()



# coding: utf-8

from struct import pack, unpack
from typing import Optional, Tuple
from time import sleep
import logging

from .meta_actuator import Actuator
from .._global import OptionalModule

try:
  import serial
except (ModuleNotFoundError, ImportError):
  serial = OptionalModule("pyserial")


cmd_header = b'\x52\x52\x52\xFF\x00'
cmd_tail = b'\xAA\xAA\x50\x50\x50\xFF\x00'
msg_tail_last = b'\xAA\xAA'
msg_tail_not_last = b'\xAA\xAA+'


class JVLMac140(Actuator):
  """This class allows driving JVL's MAC140 integrated servomotor in speed or
  in position.

  It interfaces with the servomotor over a serial connection.
  
  .. versionadded:: 1.4.0
  .. versionchanged:: 2.0.0 renamed class from Biotens to JVLMac140
  """

  def __init__(self, port: str = '/dev/ttyUSB0') -> None:
    """Initializes the parent class.

    Args:
      port: Path to the serial port to use for communication.

    .. deprecated:: 2.0.0 *baudrate* argument
    """

    self._ser = None

    super().__init__()

    self._port = port

  def open(self) -> None:
    """Initializes the serial connection and clears any serial error."""

    self.log(logging.INFO, f"Opening the serial port {self._port} with "
                           f"baudrate 19200")
    self._ser = serial.Serial(self._port, baudrate=19200, timeout=0.1)
    # Clearing any error in the motor registers
    cmd = self._make_cmd((35, 4, 0, 35), ('B', 'B', 'i', 'B'), True)
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.write(cmd)

  def set_speed(self, speed: float) -> None:
    """Sets the desired speed on the actuator.

    Args:
      speed: The target speed, in `mm/min`.
    """

    # For the conversions, there are 4096 counts/motor revolution, 1/16 encoder
    # counts/sample, and the screw thread is 5
    speed = int(round(16 * 4096 * speed / (520.8 * 60 * 5)))
    acc = int(round(16 * 4096 * 10000 / (520.8 * 520.8 * 5)))

    # Generating the commands to send
    # The torque is set to 1023, the acceleration to 10000mm/s²
    set_speed = self._make_cmd((5, 2, speed, 5), ('B', 'B', 'h', 'B'), True)
    set_torque = self._make_cmd((7, 7, 1023, 7), ('B', 'B', 'h', 'B'), True)
    set_acc = self._make_cmd((6, 2, acc, 6), ('B', 'B', 'h', 'B'), False)
    command = self._make_cmd((2, 2, 1, 2), ('B', 'B', 'h', 'B'), True)

    # Writing the command values to the motor registers
    cmd = [set_speed, set_torque, set_acc, command]
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.writelines(cmd)

  def set_position(self, position: float, speed: Optional[float]) -> None:
    """Sets the desired target position on the servomotor.

    Args:
      position: The target position, in `mm`.
      speed: The target speed for reaching the desired position, in `mm/min`.
        The speed must be given, otherwise an exception is raised.
    
    .. versionchanged:: 2.0.0 *speed* is now a mandatory argument
    """

    if speed is None:
      raise ValueError("The JVLMac140 actuator needs both a position and a "
                       "speed command when driven in position mode !")

    # For the conversions, there are 4096 counts/motor revolution, 1/16 encoder
    # counts/sample, and the screw thread is 5
    pos = int(round(position * 4096 / 5))
    speed = int(round(16 * 4096 * speed / (520.8 * 60 * 5)))
    acc = int(round(16 * 4096 * 10000 / (520.8 * 520.8 * 5)))

    # Generating the commands to send
    # The torque is set to 1023, the acceleration to 10000mm/s²
    set_position = self._make_cmd((3, 4, pos, 3), ('B', 'B', 'i', 'B'), False)
    set_speed = self._make_cmd((5, 2, speed, 5), ('B', 'B', 'h', 'B'), True)
    set_torque = self._make_cmd((7, 7, 1023, 7), ('B', 'B', 'h', 'B'), True)
    set_acc = self._make_cmd((6, 2, acc, 6), ('B', 'B', 'h', 'B'), True)
    command = self._make_cmd((2, 2, 2, 2), ('B', 'B', 'h', 'B'), True)

    # Writing the command values to the motor registers
    cmd = [set_position, set_speed, set_torque, set_acc, command]
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.writelines(cmd)

  def get_position(self) -> float:
    """Reads and returns the current position of the servomotor, in `mm`.

    .. versionchanged:: 1.5.2 renamed from get_pos to get_position
    """

    # We have 20 attempts for reading the position
    for _ in range(20):
      try:
        # Emptying the read buffer
        self._ser.readlines()
        # Sending command to return position
        cmd = b''.join((b'\x50\x50\x50\xFF\x00', self._to_bytes(10, 'B'),
                        msg_tail_last))
        self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
        self._ser.write(cmd)
        # Reading the position
        position = self._ser.read(19)
        self.log(logging.DEBUG, f"Read {position} from port {self._port}")
        # Might return fewer characters than expected due to the timeout
        if len(position) != 19:
          continue
        # Parsing the position value and returning it
        return unpack('i', position[9:17:2])[0] * 5 / 4096.
      # Catching serial errors
      except serial.SerialException:
        pass
      sleep(0.1)

    # In case no value was received after 20 attempts
    raise IOError("Could not read the position for the JVLMac140 actuator!")

  def stop(self) -> None:
    """Sends a command for stopping the servomotor."""

    if self._ser is not None:
      cmd = self._make_cmd((2, 2, 0, 2), ('B', 'B', 'h', 'B'), True)
      self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
      self._ser.write(cmd)

  def close(self) -> None:
    """Closes the serial connection to the servomotor."""

    if self._ser is not None:
      self.log(logging.INFO, f"Closing the serial port {self._port}")
      self._ser.close()

  def reset_position(self) -> None:
    """Makes the servomotor reach its limit position, in order to re-calibrate
    the position readout."""

    init_pos = self._make_cmd((38, 4, 0, 38), ('B', 'B', 'i', 'B'), True)
    init_speed = self._make_cmd((40, 2, -50, 40), ('B', 'B', 'h', 'B'), True)
    init_torque = self._make_cmd((41, 2, 1023, 41), ('B', 'B', 'i', 'B'), True)
    to_init = self._make_cmd((37, 2, 0, 37), ('B', 'B', 'h', 'B'), True)

    cmd = [init_pos, init_speed, init_torque, to_init]
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.writelines(cmd)
    cmd = self._make_cmd((2, 2, 12, 2), ('B', 'B', 'h', 'B'), True)
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.write(cmd)

    sleep(1)

    last_pos = 0
    pos = 99

    # Loop while the actuator is still moving
    while pos != last_pos:
      last_pos = pos
      pos = self.get_position()
      self.log(logging.INFO, f"Current position: {pos}")
    self.log(logging.INFO, "Initialization done")
    self.stop()

    cmd = self._make_cmd((10, 4, 0, 10), ('B', 'B', 'i', 'B'), True)
    self.log(logging.DEBUG, f"Writing {cmd} to port {self._port}")
    self._ser.write(cmd)

    # Emptying the serial read buffer
    try:
      self._ser.readlines()
    except serial.SerialException:
      pass

  def _make_cmd(self,
                values: Tuple[int, int, int, int],
                encodings: Tuple[str, str, str, str],
                last_cmd: bool) -> bytes:
    """Builds a command to send to the servomotor, from the given arguments.

    This method is meant to simplify the code in the main methods of the class.
    """

    return b''.join((cmd_header,
                     self._to_bytes(values[0], encodings[0]),
                     self._to_bytes(values[1], encodings[1]),
                     self._to_bytes(values[2], encodings[2]),
                     cmd_tail,
                     self._to_bytes(values[3], encodings[3]),
                     msg_tail_last if last_cmd else msg_tail_not_last))

  @staticmethod
  def _to_bytes(value: float, encoding: str) -> bytes:
    """Generates bytes carrying a given value with the given encoding, and
    following the correct syntax for communicating with the servomotor."""

    encoded = pack(encoding, value)
    return b''.join((bytes((enc, enc ^ 0xFF)) for enc in encoded))
